from .page_store import PageListStore
from .page_browser import PagePreviewList
from .window import MainWindow
from .dialogs import AboutDialog, OpenDialog, SaveDialog, SaveChangesDialog

__all__ = ['PageListStore', 'PagePreviewList', 'MainWindow', 'AboutDialog', 'OpenDialog', 'SaveDialog',
           'SaveChangesDialog']



from .base import View
from .html import ViewHtml
from .images import ViewImages
from .text import ViewText
from .xml import ViewXml
from .empty import ViewEmpty
from .diff import ViewDiff
from .page import ViewPage
from .registry import ViewRegistry


__all__ = ['View', 'ViewRegistry', 'ViewImages', 'ViewText', 'ViewXml', 'ViewHtml', 'ViewEmpty', 'ViewDiff', 'ViewPage']



from __future__ import annotations
from typing import TextIO, Optional, Type, TYPE_CHECKING

import sys
import os

if TYPE_CHECKING:
    from types import TracebackType


class RedirectedStdStreams:
    """
    Temporarily redirects stdout/stderr

    Usage:
    with RedirectStdStreams(stderr=open(os.devnull, 'w')):
        # do something that spams stderr

    """
    def __init__(self, stdout: TextIO = None, stderr: TextIO = None) -> None:
        self._stdout = stdout or sys.stdout
        self._stderr = stderr or sys.stderr

    def __enter__(self) -> None:
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, t: Optional[Type[BaseException]], e: Optional[BaseException], tb: TracebackType) -> None:
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr


class SilencedStreams(RedirectedStdStreams):
    """
    Temporarily redirects stdout/stderr to nothing (dev/null)

    Usage:
    with SilencedStream(False, True):
        # do something that spams stderr

    """
    def __init__(self, silence_stdout: bool = True, silence_stderr: bool = True):
        self.silence_stdout = silence_stdout
        self.silence_stderr = silence_stderr
        super().__init__(
            open(os.devnull, 'w') if self.silence_stdout else sys.stdout,
            open(os.devnull, 'w') if self.silence_stderr else sys.stderr
        )

    def __exit__(self, t: Optional[Type[BaseException]], e: Optional[BaseException], tb: TracebackType) -> None:
        super().__exit__(t, e, tb)
        if self.silence_stdout:
            self._stdout.close()
        if self.silence_stderr:
            self._stderr.close()



from gi.repository import Gtk, Pango, GObject

from typing import List, Tuple, Any, Optional, Dict, cast, TYPE_CHECKING

from math import log
from enum import Enum

from ocrd_utils.constants import MIMETYPE_PAGE, MIME_TO_EXT
from ocrd_browser.util.gtk import WhenIdle

if TYPE_CHECKING:
    from ocrd_browser.model import Document, Page


class Configurator(Gtk.Widget):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.document: Optional['Document'] = None
        self.value: Any = None

    def set_document(self, document: 'Document') -> None:
        self.document = document

    def set_page(self, page: 'Page') -> None:
        pass

    def set_value(self, value: Any) -> None:
        raise NotImplementedError('You have to override set_value')


class View:
    # TODO: Views should announce which mimetype they can handle and be only available if it there is matching mimetype in Document, also they should announce if they can handle a certain Page
    # noinspection PyTypeChecker
    def __init__(self, name: str, window: Gtk.Window):
        self.name: str = name
        self.window: Gtk.Window = window
        self.document: 'Document' = None
        self.current: Page = None
        self.page_id: str = None

        self.configurators: Dict[str, Configurator] = {}
        self.container: Gtk.Box = None
        self.action_bar: Gtk.ActionBar = None
        self.scroller: Gtk.ScrolledWindow = None

    def build(self) -> None:
        self.container = Gtk.Box(visible=True, orientation="vertical", name=self.name)
        self.action_bar = Gtk.ActionBar(visible=True)
        self.action_bar.pack_end(CloseButton(self.name))
        self.action_bar.pack_end(SplitViewButton(self.name, True))
        self.action_bar.pack_end(SplitViewButton(self.name, False))

        self.scroller = Gtk.ScrolledWindow(visible=True, shadow_type='in', propagate_natural_width=True)

        self.container.pack_start(self.action_bar, False, True, 0)
        self.container.pack_start(self.scroller, True, True, 0)

    def set_document(self, document: 'Document') -> None:
        self.document = document
        for configurator in self.configurators.values():
            configurator.set_document(document)

    def add_configurator(self, name: str, configurator: Configurator) -> None:
        """
        Adds a configurator for property self.{name}

        @param name: str
        @param configurator: Configurator|Gtk.Widget
        """
        if not hasattr(self, name):
            raise AttributeError(f'{self.__class__.__name__} has no attribute {name}.')
        configurator.set_value(getattr(self, name))
        self.configurators[name] = configurator
        configurator.connect('changed', lambda _source, *value: self.config_changed(name, value))
        self.action_bar.pack_start(configurator)

    def config_changed(self, name: str, value: Any) -> None:
        if len(value) == 1:
            value = value[0]
        setattr(self, name, value)

    @property
    def use_file_group(self) -> str:
        return 'OCR-D-IMG'

    def page_activated(self, _sender: Gtk.Widget, page_id: str) -> None:
        if page_id != self.page_id:
            self.page_id = page_id
            self.reload()

    def pages_selected(self, _sender: Gtk.Widget, page_ids: List[str]) -> None:
        pass

    def reload(self) -> None:
        if self.page_id:
            self.current = self.document.page_for_id(self.page_id, self.use_file_group)
            if self.current:
                for configurator in self.configurators.values():
                    configurator.set_page(self.current)

        WhenIdle.call(self.redraw)

    def redraw(self) -> None:
        pass

    def update_ui(self) -> None:
        pass


class PageQtySelector(Gtk.Box, Configurator):

    def __init__(self) -> None:
        super().__init__(visible=True, spacing=3)
        self.value = None

        label = Gtk.Label(label='#Pages:', visible=True)

        self.pages = Gtk.SpinButton(visible=True, max_length=2, width_chars=2, max_width_chars=2, numeric=True)
        self.pages.set_tooltip_text('How many pages should be displayed at once?')
        # noinspection PyCallByClass,PyArgumentList
        self.pages.set_adjustment(Gtk.Adjustment.new(1, 1, 16, 1, 0, 0))

        self.pack_start(label, False, True, 0)
        self.pack_start(self.pages, False, True, 0)

        self.pages.connect('value-changed', self.value_changed)

    def set_value(self, value: int) -> None:
        self.value = value
        self.pages.set_value(value)

    def value_changed(self, spin: Gtk.SpinButton) -> None:
        self.emit('changed', int(spin.get_value()))

    @GObject.Signal(arg_types=[int])
    def changed(self, page_qty: int) -> None:
        self.value = page_qty


class ImageZoomSelector(Gtk.Box, Configurator):

    def __init__(self, base: float = 2, step: float = 0.1, min_: float = -4.0, max_: float = 2.0) -> None:
        super().__init__(visible=True, spacing=2)
        self.value = None

        self.base = base
        self.scale = Gtk.SpinButton(visible=True, max_length=5, width_chars=5, max_width_chars=5, numeric=True,
                                    digits=2)
        self.scale.set_tooltip_text('Zoom: log{:+.0f} scale factor for viewing images'.format(base))
        # noinspection PyCallByClass,PyArgumentList
        self.scale.set_adjustment(Gtk.Adjustment.new(0.0, min_, max_, step, 0, 0))
        self.scale.set_snap_to_ticks(False)
        self.scale.connect('value-changed', self.value_changed)

        # label = Gtk.Label(label='Zoom:', visible=True)
        # self.pack_start(label, False, True, 0)
        self.pack_start(self.scale, False, True, 0)

    def get_exp(self) -> float:
        return cast(float, self.base ** self.value)

    def set_value(self, value: float) -> None:
        adj: Gtk.Adjustment = self.scale.get_adjustment()
        value = min(adj.get_upper(), max(adj.get_lower(), value))
        self.scale.set_value(value)

    def value_changed(self, spin: Gtk.SpinButton) -> None:
        self.emit('changed', spin.get_value())

    @GObject.Signal(arg_types=[float])
    def changed(self, scale: float) -> None:
        self.value = scale
        self.scale.set_tooltip_text(
            '{:.2%} / log{:.2f} scale factor for viewing images'.format(self.get_exp(), self.base))

    def zoom_by(self, steps: int) -> None:
        """
        Zooms in(`steps` > 0) or out(`steps` < 0) by `steps` steps
        """
        direction = Gtk.SpinType.STEP_FORWARD if steps > 0 else Gtk.SpinType.STEP_BACKWARD
        self.scale.spin(direction, abs(steps) * self.scale.get_adjustment().get_step_increment())

    def zoom_to(self, to: str, width_ratio: float, height_ratio: float) -> None:
        """
        Set zoom to one of
            'original': Original Size
            'width': Fit to width
            'height': Fit to height
            'page': Fit to show whole page
            'viewport': Fit to use whole viewport
        """
        lookup = {
            'original': lambda: 1,
            'width': lambda: width_ratio,
            'height': lambda: height_ratio,
            'page': lambda: min(width_ratio, height_ratio),
            'viewport': lambda: max(width_ratio, height_ratio)
        }
        if to in lookup:
            ratio = lookup[to]()
            self.set_value(log(ratio, self.base))
        else:
            raise ValueError('to was "{}", but needs to be one of {}'.format(to, ', '.join(lookup.keys())))


class FileGroupSelector(Gtk.Box, Configurator):

    def __init__(self, filter_: Optional['FileGroupFilter'] = None, show_mime: bool = False, show_ext: bool = True):
        super().__init__(visible=True, spacing=2)
        self.value = None

        self.groups = FileGroupComboBox(filter_, show_mime, show_ext)

        # label = Gtk.Label(label='Group:', visible=True)
        # self.pack_start(label, False, True, 0)
        self.pack_start(self.groups, False, True, 0)

        self.groups.connect('changed', self.combo_box_changed)

    def set_value(self, value: Tuple[str, str]) -> None:
        self.value = value
        active_id = None
        for id_, group, mime, _ext in self.groups.get_model():
            if (value[0] is None or value[0] == group) and (value[1] is None or value[1] == mime):
                active_id = id_
                break
        if active_id:
            self.groups.set_active_id(active_id)

    def set_document(self, document: 'Document') -> None:
        self.groups.set_document(document)
        self.set_value(self.value)

    def combo_box_changed(self, combo: Gtk.ComboBox) -> None:
        model = combo.get_model()
        if len(model) > 0:
            row = combo.get_model()[combo.get_active()][:]
            self.emit('changed', row[FileGroupComboBox.COLUMN_GROUP], row[FileGroupComboBox.COLUMN_MIME])

    @GObject.Signal()
    def changed(self, file_group: str, mime_type: str) -> None:
        self.value = (file_group, mime_type)


class FileGroupComboBox(Gtk.ComboBox):
    COLUMN_ID = 0
    COLUMN_GROUP = 1
    COLUMN_MIME = 2
    COLUMN_EXT = 3

    def __init__(self, filter_: Optional['FileGroupFilter'] = None, show_mime: bool = False, show_ext: bool = True):
        Gtk.ComboBox.__init__(self, visible=True, has_tooltip=True, popup_fixed_width=False)
        self.set_model(Gtk.ListStore(str, str, str, str))
        self.filter = filter_
        self.set_id_column(self.COLUMN_ID)

        self.add_renderer(self.COLUMN_GROUP)
        if show_mime:
            self.add_renderer(self.COLUMN_MIME, 80)
        if show_ext:
            self.add_renderer(self.COLUMN_EXT)

        self.connect('query-tooltip', self.set_tooltip)

    def set_tooltip(self, _widget: Gtk.Widget, _x: int, _y: int, _keyboard_mode: bool, tooltip: Gtk.Tooltip) -> bool:
        model = self.get_model()
        if len(model) > 0:
            row = self.get_model()[self.get_active()][:]
            phs = {'fileGrp': row[self.COLUMN_GROUP], 'ext': row[self.COLUMN_EXT], 'mime': row[self.COLUMN_MIME]}
            tooltip.set_text('fileGrp: {fileGrp} (Mime-Type: {mime})'.format(**phs))
            return True
        return False

    def set_document(self, document: 'Document') -> None:
        self.set_model(FileGroupModel.build(document, self.filter))
        # self.set_active(0)

    def add_renderer(self, column: int, width: int = None) -> Gtk.CellRendererText:
        renderer = Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.MIDDLE)
        if width:
            renderer.props.width = width
        self.pack_start(renderer, False)
        self.add_attribute(renderer, "text", column)
        return renderer


class FileGroupModel(Gtk.ListStore):
    def __init__(self, document: 'Document'):
        super().__init__(str, str, str, str)
        for group, mime in document.file_groups_and_mimetypes:
            if mime == 'text/html':
                ext = '.html'
            else:
                ext = MIME_TO_EXT.get(mime, '.???')
            self.append(('{}|{}'.format(group, mime), group, mime, ext))

    @classmethod
    def build(cls, document: 'Document', filter_: 'FileGroupFilter' = None) -> 'FileGroupModel':
        model = cls(document)
        model = model.filter_new()
        model.set_visible_func(filter_ if filter_ else FileGroupFilter.ALL)
        return model

    @staticmethod
    def image_filter(model: Gtk.TreeModel, it: Gtk.TreeIter, _data: None) -> bool:
        return str(model[it][FileGroupComboBox.COLUMN_MIME]).startswith('image/')

    @staticmethod
    def page_filter(model: Gtk.TreeModel, it: Gtk.TreeIter, _data: None) -> bool:
        return str(model[it][FileGroupComboBox.COLUMN_MIME]) == str(MIMETYPE_PAGE)

    @staticmethod
    def xml_filter(model: Gtk.TreeModel, it: Gtk.TreeIter, _data: None) -> bool:
        return str(model[it][FileGroupComboBox.COLUMN_EXT]) == '.xml'

    @staticmethod
    def html_filter(model: Gtk.TreeModel, it: Gtk.TreeIter, _data: None) -> bool:
        # str casts for mypy
        return str(model[it][FileGroupComboBox.COLUMN_MIME]) == 'text/html'

    @staticmethod
    def all_filter(_model: Gtk.TreeModel, _it: Gtk.TreeIter, _data: None) -> bool:
        return True


class FileGroupFilter(Enum):
    IMAGE = FileGroupModel.image_filter
    PAGE = FileGroupModel.page_filter
    HTML = FileGroupModel.html_filter
    XML = FileGroupModel.xml_filter
    ALL = FileGroupModel.all_filter


class CloseButton(Gtk.Button):
    def __init__(self, view_name: str):
        Gtk.Button.__init__(self, visible=True)
        self.set_name('close_{}'.format(view_name))
        self.set_detailed_action_name('win.close_view("{}")'.format(view_name))
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.set_always_show_image(True)
        # noinspection PyArgumentList
        self.set_image(Gtk.Image.new_from_icon_name('window-close-symbolic', Gtk.IconSize.SMALL_TOOLBAR))


class SplitViewButton(Gtk.Button):
    def __init__(self, view_name: str, vertical: bool = True):
        Gtk.Button.__init__(self, visible=True)
        direction = 'vertical' if vertical else 'horizontal'
        self.set_name('split_{}_{}'.format(view_name, direction))
        self.set_detailed_action_name('win.split_view(("{}","empty", {}))'.format(view_name, 'true' if vertical else 'false'))
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.set_always_show_image(True)
        # noinspection PyArgumentList
        self.set_image(Gtk.Image.new_from_icon_name('split-{}'.format(direction), Gtk.IconSize.SMALL_TOOLBAR))



from tests import TestCase, TEST_BASE_PATH, ASSETS_PATH
from ocrd_browser.model import Document, IMAGE_FROM_PAGE_FILENAME_SUPPORT


class PageTestCase(TestCase):

    def test_xpath_works_with_different_namespaces(self):
        doc = Document.load(TEST_BASE_PATH / 'example/workspaces/aletheiaexamplepage/mets.xml')
        for page_id in ['PAGE_2017', 'PAGE_2018', 'PAGE_2019']:
            page = doc.page_for_id(page_id, 'OCR-D-GT-PAGE')
            xpath_result = page.xpath('/page:PcGts/page:Page/@imageFilename')
            self.assertGreater(len(xpath_result), 0)

    def test_can_call_get_image_if_supported(self):
        page = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784-binarized/data/mets.xml').page_for_id('P_0017', 'OCR-D-GT-WORD')
        if IMAGE_FROM_PAGE_FILENAME_SUPPORT:
            image_by_feature, _, _ = page.get_image(feature_selector={'binarized'}, feature_filter={'cropped'})
            image_by_filename, _, _ = page.get_image(filename='OCR-D-IMG-BIN/BIN_0017.png', feature_filter={'cropped'})
            self.assertEqual(image_by_feature, image_by_filename)
        else:
            try:
                page.get_image(filename='OCR-D-GT-IMG-BIN/PAGE_2019.tif')
                self.fail('IMAGE_FROM_PAGE_FILENAME_SUPPORT detected wrong')
            except RuntimeError as e:
                self.assertTrue(str(e).startswith('Parameter filename not supported in '))

    def test_missing_image(self):
        doc = Document.load((TEST_BASE_PATH / 'example/workspaces/kant_aufklaerung_1784_missing_image/mets.xml').as_uri())
        page = doc.page_for_id('PHYS_0017', 'OCR-D-GT-PAGE')
        image, info, exif = page.get_image(feature_selector='', feature_filter='binarized')
        # Assert no exceptions happened but image is None
        self.assertIsNone(image)



from typing import Optional

from gi.repository import Gtk

from ocrd_browser.view import View


class ViewEmpty(View):
    """
    A view of the current PAGE-XML TextEquiv annotation concatenated as plain text
    """
    label: Gtk.Label

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.viewport: Optional[Gtk.Viewport] = None
        self.box: Optional[Gtk.Box] = None

    def build(self) -> None:
        super().build()
        self.viewport = Gtk.Viewport(visible=True, hscroll_policy='natural', vscroll_policy='natural',
                                     parent=self.scroller)
        self.box = Gtk.Box(visible=True, parent=self.viewport, orientation=Gtk.Orientation.VERTICAL)
        self.label: Gtk.Label = Gtk.Label(visible=True, vexpand=False)
        self.label.set_text('Create New View:')
        self.box.pack_start(self.label, False, False, 0)
        for id_, view in self.window.view_registry.get_view_options().items():
            menu_item = Gtk.ModelButton(visible=True, centered=False, halign=Gtk.Align.FILL, label=view, hexpand=True)
            menu_item.set_detailed_action_name('win.replace_view(("{}", "{}"))'.format(self.name, id_))
            self.box.pack_start(menu_item, False, False, 0)



import unittest

from ocrd_browser.util.config import Settings
from tests import TestCase, TEST_BASE_PATH


class SettingsTestCase(TestCase):

    def setUp(self) -> None:
        self.settings = Settings.build_from_files([TEST_BASE_PATH / 'example/config/simple.conf'], validate=False)

    def test_default_value(self):
        settings = Settings({})
        self.assertEqual(['OCR-D-IMG', 'OCR-D-IMG.*'], settings.file_groups.preferred_images)

    def test_loaded_value(self):
        self.assertEqual(['OCR-D-IMG', 'OCR-D-IMG.*', 'ORIGINAL'], self.settings.file_groups.preferred_images)

    def test_tools(self):
        pv = self.settings.tools['PageViewer']
        self.assertEqual(
            '/usr/bin/java -jar /home/jk/bin/JPageViewer/JPageViewer.jar --resolve-dir {workspace.directory} {file.path.absolute}',
            pv.commandline)
        self.assertEqual('p', pv.shortcut)

        pv = self.settings.tools['Open']
        self.assertEqual('xdg-open {file.path.absolute}', pv.commandline)
        self.assertEqual('o', pv.shortcut)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- Mode: Python; coding: utf-8; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-
import os
import sys
import pstats
import io
import cProfile

import gi

gi.require_version('Gtk', '3.0')
gi.require_version('GdkPixbuf', '2.0')
try:
    gi.require_version('GtkSource', '4')
except ValueError:
    gi.require_version('GtkSource', '3.0')

gi.require_version('WebKit2', '4.0')

from gi.repository import Gtk, Gio, GLib  # noqa: E402
from pathlib import Path  # noqa: E402
from typing import Type  # noqa: E402
from types import TracebackType  # noqa: E402

BASE_PATH = Path(__file__).absolute().parent
resources = Gio.resource_load(str(BASE_PATH / "ui.gresource"))
Gio.resources_register(resources)

PROFILER = None
if 'STARTUP_PROFILE' in os.environ:
    PROFILER = cProfile.Profile()
    PROFILER.enable()


def install_excepthook() -> None:
    """ Make sure we exit when an unhandled exception occurs. """
    old_hook = sys.excepthook

    def new_hook(type_: Type[BaseException], value: BaseException, traceback: TracebackType) -> None:
        old_hook(type_, value, traceback)
        while Gtk.main_level():
            Gtk.main_quit()
        sys.exit()

    sys.excepthook = new_hook


def startup_time() -> None:
    PROFILER.disable()
    s = io.StringIO()
    ps = pstats.Stats(PROFILER, stream=s).sort_stats(pstats.SortKey.TIME)
    ps.print_stats(20)
    print(s.getvalue())


def main() -> None:
    if PROFILER:
        GLib.idle_add(startup_time)
    from ocrd_utils import initLogging
    initLogging()
    from ocrd_browser.application import OcrdBrowserApplication
    install_excepthook()
    app = OcrdBrowserApplication()
    app.run(sys.argv)


if __name__ == "__main__":
    # WHY OH WHY
    sys.path.append(str(BASE_PATH.parent))
    main()
    sys.exit()



from gi.repository import Gtk, GdkPixbuf, Gio, GObject, GLib, Gdk
from ocrd_models import OcrdFile

from ocrd_browser.model import Document
from ocrd_browser.view import ViewRegistry, ViewPage
from ocrd_browser.util.gtk import ActionRegistry, resource_string
from .dialogs import SaveDialog, SaveChangesDialog
from .page_browser import PagePreviewList
from typing import List, cast, Any, Optional

from ..view.manager import ViewManager


@Gtk.Template(string=resource_string('main-window.ui'))
class MainWindow(Gtk.ApplicationWindow):
    __gtype_name__ = "MainWindow"

    header_bar: Gtk.HeaderBar = Gtk.Template.Child()
    page_list_scroller: Gtk.ScrolledWindow = Gtk.Template.Child()
    current_page_label: Gtk.Label = Gtk.Template.Child()
    view_container: Gtk.Box = Gtk.Template.Child()
    view_menu_box: Gtk.Box = Gtk.Template.Child()

    def __init__(self, **kwargs: Any):
        Gtk.ApplicationWindow.__init__(self, **kwargs)
        # noinspection PyCallByClass,PyArgumentList
        self.set_icon(GdkPixbuf.Pixbuf.new_from_resource("/org/readmachine/ocrd-browser/icons/icon.png"))
        self.view_manager = ViewManager(self, self.view_container)
        self.current_page_id: Optional[str] = None
        # noinspection PyTypeChecker
        self.document = Document.create(emitter=self.emit)

        self.actions = ActionRegistry(for_widget=self)
        self.actions.create('close')
        self.actions.create('goto_first')
        self.actions.create('go_back')
        self.actions.create('go_forward')
        self.actions.create('goto_last')
        self.actions.create('page_remove')
        self.actions.create('page_properties')
        self.actions.create('close_view', param_type=GLib.VariantType("s"))
        self.actions.create('split_view', param_type=GLib.VariantType("(ssb)"))
        self.actions.create('create_view', param_type=GLib.VariantType("s"))
        self.actions.create('replace_view', param_type=GLib.VariantType("(ss)"))
        self.actions.create('toggle_edit_mode', state=GLib.Variant('b', False))
        self.actions.create('save')
        self.actions.create('save_as')

        self.connect('delete-event', self.on_delete_event)

        self.page_list = PagePreviewList(self.document)
        self.page_list_scroller.add(self.page_list)
        self.page_list.connect('page_activated', self.on_page_activated)
        self.page_list.connect('pages_selected', self.on_pages_selected)

        for id_, view in self.view_registry.get_view_options().items():
            menu_item = Gtk.ModelButton(visible=True, centered=False, halign=Gtk.Align.FILL, label=view, hexpand=True)
            menu_item.set_detailed_action_name('win.create_view("{}")'.format(id_))
            self.view_menu_box.pack_start(menu_item, True, True, 0)

        self.view_manager.set_root_view(ViewPage)
        # self.view_manager.split(None, ViewPage, False)

        self.update_ui()

    def on_page_remove(self, _a: Gio.SimpleAction, _p: None) -> None:
        for page_id in self.page_list.get_selected_ids():
            self.document.delete_page(page_id)

    def on_page_properties(self, _a: Gio.SimpleAction, _p: None) -> None:
        pass

    @Gtk.Template.Callback()
    def on_recent_menu_item_activated(self, recent_chooser: Gtk.RecentChooserMenu) -> None:
        app = self.get_application()
        item: Gtk.RecentInfo = recent_chooser.get_current_item()
        app.open_in_window(item.get_uri(), window=self)

    def open(self, uri: str) -> None:
        self.set_title('Loading')
        self.header_bar.set_title('Loading ...')
        self.header_bar.set_subtitle(uri)
        self.update_ui()
        # Give GTK some break to show the Loading message
        GLib.timeout_add(50, self._open, uri)

    def _open(self, uri: str) -> None:
        # noinspection PyTypeChecker
        self.document = Document.load(uri, emitter=self.emit)
        self.page_list.set_document(self.document)

        self.view_manager.set_document(self.document)
        self.update_ui()

        if len(self.document.page_ids):
            self.on_page_activated(None, self.document.page_ids[0])

    @property
    def view_registry(self) -> ViewRegistry:
        return cast(MainWindow, self.get_application()).view_registry

    def on_page_activated(self, _sender: Optional[Gtk.Widget], page_id: str) -> None:
        if self.current_page_id != page_id:
            self.current_page_id = page_id
            self.emit('page_activated', page_id)

    @GObject.Signal(arg_types=[str])
    def page_activated(self, page_id: str) -> None:
        index = self.document.page_ids.index(page_id)
        self.current_page_label.set_text('{}/{}'.format(index + 1, len(self.document.page_ids)))
        self.update_ui()

    def on_pages_selected(self, _sender: Optional[Gtk.Widget], page_ids: List[str]) -> None:
        self.emit('pages_selected', page_ids)

    @GObject.Signal(arg_types=[object])
    def pages_selected(self, page_ids: List[str]) -> None:
        pass

    @GObject.Signal(arg_types=[str, object])
    def document_changed(self, subtype: str, page_ids: List[str]) -> None:
        self.page_list.document_changed(subtype, page_ids)
        self.update_ui()

    @GObject.Signal(arg_types=[object])
    def document_saved(self, _saved: Document) -> None:
        self.update_ui()

    @GObject.Signal(arg_types=[float, object])
    def document_saving(self, progress: float, file: Optional[OcrdFile]) -> None:
        pass

    def update_ui(self) -> None:
        title = self.document.title + (' *' if self.document.modified else '')
        self.set_title(title)
        self.header_bar.set_title(title)
        self.header_bar.set_subtitle(self.document.original_url)

        can_go_back = False
        can_go_forward = False
        if self.current_page_id and self.current_page_id in self.document.page_ids:
            index = self.document.page_ids.index(self.current_page_id)
            last_page = len(self.document.page_ids) - 1
            can_go_back = index > 0
            can_go_forward = index < last_page

        self.actions['goto_first'].set_enabled(can_go_back)
        self.actions['go_back'].set_enabled(can_go_back)
        self.actions['go_forward'].set_enabled(can_go_forward)
        self.actions['goto_last'].set_enabled(can_go_forward)
        self.actions['page_remove'].set_enabled(self.document.editable)
        # noinspection PyCallByClass,PyArgumentList
        self.actions['toggle_edit_mode'].set_state(GLib.Variant.new_boolean(self.document.editable))
        self.actions['save'].set_enabled(self.document.modified)
        self.view_manager.update_ui()

    def close_confirm(self) -> bool:
        if self.document.modified:
            # Do you wanna save the changes?
            d = SaveChangesDialog(document=self.document, transient_for=self, modal=True)
            save_changes = d.run()
            d.destroy()

            if save_changes == Gtk.ResponseType.NO:
                return True
            elif save_changes == Gtk.ResponseType.CANCEL:
                return False
            elif save_changes == Gtk.ResponseType.YES:
                return self.on_save()
            return False
        else:
            return True

    def on_close(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        if self.close_confirm():
            self.destroy()

    def on_delete_event(self, _window: 'MainWindow', _event: Gdk.Event) -> bool:
        return not self.close_confirm()

    def on_goto_first(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.page_list.goto_index(0)

    def on_go_forward(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.page_list.skip(1)

    def on_go_back(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.page_list.skip(-1)

    def on_goto_last(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.page_list.goto_index(-1)

    def on_create_view(self, _a: Gio.SimpleAction, selected_view_id: GLib.Variant) -> None:
        view_class = self.view_registry.get_view(selected_view_id.get_string())
        self.view_manager.add(view_class)

    def on_replace_view(self, _a: Gio.SimpleAction, arguments: GLib.Variant) -> None:
        (replace_view, new_view_name) = arguments
        new_view_type = self.view_registry.get_view(new_view_name)
        self.view_manager.replace(replace_view, new_view_type)

    def on_close_view(self, _action: Gio.SimpleAction, view_name: GLib.Variant) -> None:
        try:
            self.view_manager.close(view_name.get_string())
        except ValueError:
            # Tried to remove last view
            pass

    def on_split_view(self, _action: Gio.SimpleAction, arguments: GLib.Variant) -> None:
        (split_view, new_view_name, horizontal) = arguments
        new_view_type = self.view_registry.get_view(new_view_name)
        self.view_manager.split(split_view, new_view_type, horizontal)

    def on_save(self, _a: Gio.SimpleAction = None, _p: None = None) -> bool:
        if self.document.original_url:
            return self.save()
        else:
            return self.save_as()

    def on_save_as(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.save_as()

    def save(self) -> bool:
        self.document.save()
        self.update_ui()
        return True

    def save_as(self) -> bool:
        save_dialog = SaveDialog(application=self.get_application(), transient_for=self, modal=True)
        if self.document.original_url:
            save_dialog.set_filename(self.document.original_url)
        else:
            save_dialog.set_current_name('mets.xml')
        response = save_dialog.run()
        should_save = bool(response == Gtk.ResponseType.OK)
        if should_save:
            self.document.save_as(save_dialog.get_uri())
        save_dialog.destroy()
        self.update_ui()
        return should_save

    def on_toggle_edit_mode(self, _a: Gio.SimpleAction = None, _p: None = None) -> None:
        self.document.editable = not self.document.editable
        self.update_ui()



from gi.repository import Gtk, Gdk, GObject, Pango, Gio, GLib

from typing import Any, Optional, Tuple, Dict, List, NamedTuple, FrozenSet

from pathlib import Path
from PIL import Image
from cairo import Context
from xml.sax.saxutils import escape

from ocrd_models.ocrd_page import AlternativeImageType
from shapely.geometry import Polygon

from ocrd_browser.util.image import pil_to_pixbuf, pil_scale
from ocrd_utils.constants import MIMETYPE_PAGE
from .base import (
    View,
    FileGroupSelector,
    FileGroupFilter,
    ImageZoomSelector,
    Configurator
)
from ..model import Page, Document, IMAGE_FROM_PAGE_FILENAME_SUPPORT
from ..model.page_xml_renderer import PageXmlRenderer, RegionMap, Feature, Region
from ..util.gtk import WhenIdle, ActionRegistry


class FeatureDescription:
    def __init__(self, icon: str, label: str, xpath: str, tooltip: str = None):
        self.icon = icon
        self.label = label
        self.xpath = xpath
        self.tooltip = tooltip if tooltip else 'Show ' + label

    def available(self, page: Page) -> bool:
        return len(page.xpath(self.xpath)) > 0


def clamp(x: float, lower: float, upper: float) -> float:
    return lower if x < lower else upper if x > upper else x


class Transformation:
    """
    Encapsulates forward and inverse affine transform consisting of scaling and translation only
    """
    def __init__(self, scale: float, tx: float, ty: float, cx: float, cy: float):
        self.scale = scale
        self.tx = tx
        self.ty = ty
        self.cx = cx
        self.cy = cy

    def transform_region(self, poly: Polygon) -> Tuple[float, float, float, float]:
        (l, t), (r, b) = self.inverse(*poly.bounds[0:2]), self.inverse(*poly.bounds[2:4])
        return l, t, r - l, b - t

    def transform(self, x: float, y: float) -> Tuple[float, float]:
        return clamp((x + self.tx) * self.scale, 0, self.cx), clamp((y + self.ty) * self.scale, 0, self.cy)

    def inverse(self, x: float, y: float) -> Tuple[float, float]:
        return clamp(x, 0, self.cx) / self.scale - self.tx, clamp(y, 0, self.cy) / self.scale - self.ty


class ImageFeatures:
    FEATURES = {
        'binarized': 'B',
        'grayscale_normalized': 'N',
        'deskewed': 'S',
        'despeckled': 'P',
        'dewarped': 'W',
        'cropped': 'C',
        'rotated-90': '',
        'rotated-180': '',
        'rotated-270': '',
    }

    @classmethod
    def allowed(cls) -> FrozenSet[str]:
        return frozenset(cls.FEATURES.keys())

    @classmethod
    def from_string(cls, string: str) -> FrozenSet[str]:
        return frozenset(str(string).split(',')).intersection(cls.allowed())

    @classmethod
    def negate(cls, st: FrozenSet[str]) -> FrozenSet[str]:
        return cls.allowed().difference(st.intersection(cls.allowed()))

    @classmethod
    def short(cls, st: FrozenSet[str]) -> str:
        return ''.join([cls.FEATURES[i] for i in st if i in cls.FEATURES])


class ImageVersion(NamedTuple):
    path: Path
    size: Tuple[int, int]
    features: FrozenSet[str] = frozenset()
    conf: Optional[float] = None

    def as_row(self) -> Tuple[str, str, str, str, str]:
        return ','.join(sorted(self.features)), '{0:d}✕{1:d}'.format(*self.size), ImageFeatures.short(self.features), str(self.path), self.path.parent.name

    @classmethod
    def list_from_page(cls, doc: Document, page: Page) -> List['ImageVersion']:
        versions = []
        if page:
            path = doc.path(page.page.imageFilename)
            if path.exists():
                versions.append(cls(path.relative_to(doc.directory), (page.page.imageWidth, page.page.imageHeight), frozenset(), None))
            alts: List[AlternativeImageType] = page.page.get_AlternativeImage()
            for alt in alts:
                path = doc.path(alt.filename)
                if path.exists():
                    versions.append(cls(path.relative_to(doc.directory), Image.open(path).size, frozenset(str(alt.comments).split(',')), float(alt.conf) if alt.conf is not None else None))
        return versions


class ImageVersionSelector(Gtk.Box, Configurator):

    COLUMN_FEATURES = 0
    COLUMN_SIZE = 1
    COLUMN_SHORT_FEATURES = 2
    COLUMN_PATH = 3
    COLUMN_FILE_GROUP = 4

    def __init__(self) -> None:
        super().__init__(visible=True, spacing=3)
        self.value = None

        self.versions = Gtk.ListStore(str, str, str, str, str)
        self.version_box = Gtk.ComboBox(visible=True, model=self.versions)

        self.version_box.set_id_column(self.COLUMN_PATH)

        # label = Gtk.Label(label='Image:', visible=True)
        # self.pack_start(label, False, True, 0)
        self.pack_start(self.version_box, False, True, 0)

        renderer = Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.START)
        self.version_box.pack_start(renderer, False)
        self.version_box.add_attribute(renderer, "text", self.COLUMN_FILE_GROUP)
        self.set_tooltip_text('Image-Version')

        renderer = Gtk.CellRendererText()
        self.version_box.pack_start(renderer, False)
        self.version_box.add_attribute(renderer, "text", self.COLUMN_SHORT_FEATURES)

        self._change_handler = self.version_box.connect('changed', self.combo_box_changed)

        self.connect('query-tooltip', self.set_tooltip)

    def set_tooltip(self, _widget: Gtk.Widget, _x: int, _y: int, _keyboard_mode: bool, tooltip: Gtk.Tooltip) -> bool:
        model = self.versions
        if len(model) > 0:
            row = model[self.version_box.get_active()][:]
            phs = {
                'features': row[self.COLUMN_FEATURES],
                'size': row[self.COLUMN_SIZE],
                'short_features': row[self.COLUMN_SHORT_FEATURES],
                'path': row[self.COLUMN_PATH],
                'file_group': row[self.COLUMN_FILE_GROUP],
            }
            tooltip.set_text('{path}\t {size}\t{features}'.format(**phs))
            return True
        return False

    def set_value(self, value: Tuple[str, str]) -> None:
        self.value = value
        if value and value[0]:
            self.version_box.set_active_id(value[0])

    @GObject.Signal()
    def changed(self, path: str, features: str) -> None:
        self.value = (path, features)

    def set_page(self, page: Page) -> None:
        previous_selected = None, None
        if self.version_box.get_active() != -1:
            row = self.versions[self.version_box.get_active()]
            previous_selected = row[self.COLUMN_FILE_GROUP], row[self.COLUMN_FEATURES]

        with self.version_box.handler_block(self._change_handler):
            self.versions.clear()
            for version in ImageVersion.list_from_page(self.document, page):
                self.versions.append(version.as_row())

        if self.value is None or self.value[0] is None:
            self.version_box.set_active(0)
        else:
            if self.value[0] != self.version_box.get_active_id():
                for v in self.versions:
                    if v[self.COLUMN_FILE_GROUP] == previous_selected[0] and v[self.COLUMN_FEATURES] == previous_selected[1]:
                        self.version_box.set_active_id(v[self.COLUMN_PATH])
                        return
            self.version_box.set_active_id(self.value[0])

    def combo_box_changed(self, combo: Gtk.ComboBox) -> None:
        if len(self.versions) > 0:
            row = self.versions[combo.get_active()][:]
            self.emit('changed', row[self.COLUMN_PATH], row[self.COLUMN_FEATURES])


class PageFeaturesSelector(Gtk.Box, Configurator):

    FEATURES: Dict[Feature, FeatureDescription] = {
        Feature.IMAGE: FeatureDescription('🖺', 'image', '/page:PcGts/page:Page/@imageFilename'),
        Feature.BORDER: FeatureDescription('🗋', 'border', '/page:PcGts/page:Page/page:Border/page:Coords'),
        Feature.PRINT_SPACE: FeatureDescription('🗌', 'printspace', '/page:PcGts/page:Page/page:PrintSpace/page:Coords'),
        Feature.ORDER: FeatureDescription('↯', 'order', '/page:PcGts/page:Page/page:ReadingOrder/*'),
        Feature.REGIONS: FeatureDescription('⬓', 'regions', '/page:PcGts/page:Page/*[not(local-name(.) = "Border" or local-name(.) = "PrintSpace")]/page:Coords'),
        Feature.LINES: FeatureDescription('𝌆', 'lines', '/page:PcGts/page:Page/*//page:TextLine/page:Coords'),
        Feature.BASELINES: FeatureDescription('–', 'baselines', '/page:PcGts/page:Page/*//page:TextLine/page:Baseline'),
        Feature.WORDS: FeatureDescription('𝌶', 'words', '/page:PcGts/page:Page/*//page:TextLine/page:Word/page:Coords'),
        Feature.GLYPHS: FeatureDescription('𝍖', 'glyphs', '/page:PcGts/page:Page/*//page:TextLine/page:Word/page:Glyph/page:Coords'),
        # Feature.GRAPHEMES: FeatureDescription('#', 'graphemes', '/page:PcGts/page:Page/*//page:TextLine/page:Word/page:Glyph/page:Graphemes/*/page:Coords')
        Feature.WARNINGS: FeatureDescription('‽', 'warnings', '/page:PcGts/page:Page//page:Coords', 'Mark regions with warnings in red')
    }

    def __init__(self) -> None:
        super().__init__(visible=True, spacing=3)
        self.value = None
        self.items: Dict[Feature, Gtk.CheckMenuItem] = {}

        menubutton = Gtk.MenuButton(label='Show:', visible=True)
        self.menu_label: Gtk.AccelLabel = menubutton.get_child()

        menu = Gtk.Menu(visible=True)
        for feature, desc in self.FEATURES.items():
            item = Gtk.CheckMenuItem(label=desc.label, name=feature.name, tooltip_text=desc.tooltip, visible=True)
            item.get_child().set_markup('<tt>{0.icon}</tt>\t{0.label}'.format(desc))
            item.connect("toggled", self.on_feature_toggled)
            self.items[feature] = item
            menu.append(item)

        menubutton.set_direction(Gtk.ArrowType.DOWN)
        menubutton.set_popup(menu)

        self.pack_start(menubutton, False, False, 0)

    def on_feature_toggled(self, item: Gtk.CheckMenuItem) -> None:
        toggle_feature: Feature = Feature[item.get_name()]
        if item.get_active():
            self.value = Feature(self.value) | toggle_feature
        else:
            self.value = Feature(self.value) & ~toggle_feature
        self.emit('changed', self.value)

    def update_label_markup(self) -> None:
        markup = 'Show: '
        for feature, desc in self.FEATURES.items():
            s = '<tt>' + desc.icon + '</tt>'
            if not self.items[feature].get_sensitive():
                s = '<span background="#eeeeee" foreground="#bbbbbb">{}</span>'.format(s)
            elif self.items[feature].get_active():
                s = '<span background="#ffffff" foreground="#000000"><b>{}</b></span>'.format(s)
            else:
                s = '<span background="#ffffff" foreground="#999999">{}</span>'.format(s)

            markup += s
        self.menu_label.set_markup(markup)

    def set_value(self, value: int) -> None:
        self.value = Feature(value)
        for feature in self.FEATURES:
            self.items[feature].set_active(feature & value)
        self.emit('changed', self.value)

    @GObject.Signal(arg_types=[object])
    def changed(self, features: Feature) -> None:
        self.update_label_markup()
        self.value = features

    def set_page(self, page: Page) -> None:
        if page:
            for feature, desc in self.FEATURES.items():
                self.items[feature].set_sensitive(desc.available(page))
            self.update_label_markup()


class ViewPage(View):
    """
    PageViewer like View
    """

    label = 'Page'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.current: Optional[Page] = None

        # Configurators
        self.file_group: Tuple[Optional[str], Optional[str]] = (None, MIMETYPE_PAGE)
        self.scale: float = -2.0
        self.image_version: Tuple[Optional[str], str] = (None, '')
        self.features: Feature = Feature.DEFAULT

        # GTK
        self.image: Optional[Gtk.Image] = None
        self.highlight: Optional[Gtk.DrawingArea] = None
        self.status_bar: Optional[Gtk.Box] = None

        # Data
        self.page_image: Optional[Image.Image] = None
        self.region_map: Optional[RegionMap] = None
        self.t: Optional[Transformation] = None
        self.current_region: Optional[Region] = None
        self.last_rescale: int = -100
        self.viewport_size: Gdk.Rectangle

    def build(self) -> None:
        super(ViewPage, self).build()

        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.PAGE))
        self.add_configurator('scale', ImageZoomSelector(2.0, 0.05, -4.0, 2.0))
        self.add_configurator('image_version', ImageVersionSelector())
        self.add_configurator('features', PageFeaturesSelector())
        icon = Gtk.Image.new_from_icon_name('camera-photo', Gtk.IconSize.SMALL_TOOLBAR)
        button = Gtk.Button(image=icon, visible=True, always_show_image=True, tooltip_text='Saves a screenshot of the current view')
        button.connect('clicked', self.open_screenshotdialog)
        self.action_bar.pack_start(button)

        actions = ActionRegistry()
        actions.create(name='zoom_by', param_type=GLib.VariantType('i'), callback=self._on_zoom_by)
        actions.create(name='zoom_to', param_type=GLib.VariantType('s'), callback=self._on_zoom_to)

        self.image = Gtk.Image(visible=True, icon_name='gtk-missing-image', icon_size=Gtk.IconSize.DIALOG, valign=Gtk.Align.START)

        self.highlight = Gtk.DrawingArea(visible=True, valign=Gtk.Align.FILL, halign=Gtk.Align.FILL, can_focus=True, has_focus=True, focus_on_click=True, is_focus=True)
        self.highlight.add_events(Gdk.EventMask.SMOOTH_SCROLL_MASK | Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.POINTER_MOTION_MASK)
        self.highlight.set_has_tooltip(True)
        self.highlight.connect('query-tooltip', self._query_tooltip)
        self.highlight.connect('scroll-event', self._on_scroll)
        self.highlight.connect('button-press-event', self._on_button)
        self.highlight.connect('motion-notify-event', self._on_mouse)
        self.highlight.connect('draw', self.draw_highlight)
        self.highlight.insert_action_group("view", actions.for_widget)

        overlay = Gtk.Overlay(visible=True)
        overlay.add(self.image)
        overlay.add_overlay(self.highlight)

        viewport = Gtk.Viewport(visible=True, hscroll_policy='natural', vscroll_policy='natural')
        viewport.connect('size-allocate', self._on_viewport_size_allocate)
        viewport.add(overlay)

        self.scroller.add(viewport)
        self.status_bar = Gtk.Box(visible=True, orientation=Gtk.Orientation.HORIZONTAL)
        self.container.pack_end(self.status_bar, False, False, 0)
        self.update_status_bar()

    def config_changed(self, name: str, value: Any) -> None:
        super(ViewPage, self).config_changed(name, value)
        if name == 'file_group':
            WhenIdle.call(self.reload, priority=10)
        if name == 'features' or name == 'image_version':
            WhenIdle.call(self.redraw, priority=50)
        if name == 'scale':
            WhenIdle.call(self.rescale)

    @property
    def use_file_group(self) -> str:
        return self.file_group[0]

    def redraw(self) -> None:
        got_result = False
        if self.current:
            selected_features = ImageFeatures.from_string(self.image_version[1])
            parameters = {
                'feature_selector': ','.join(selected_features),
                'feature_filter': ','.join(ImageFeatures.negate(selected_features))
            }
            if IMAGE_FROM_PAGE_FILENAME_SUPPORT:
                parameters['filename'] = self.image_version[0]
            page_image, page_coords, _ = self.current.get_image(**parameters)
            if page_image:
                renderer = PageXmlRenderer(page_image, page_coords, self.current.id, self.features)
                renderer.render_all(self.current.pc_gts)
                self.page_image, self.region_map = renderer.get_result()
                self.current_region = self.region_map.refetch(self.current_region)
                got_result = True
        if not got_result:
            self.page_image, self.region_map = None, None
        self.update_transformation()
        WhenIdle.call(self.rescale, force=True)

    def rescale(self, force: bool = False) -> None:
        if self.page_image:
            scale_config: ImageZoomSelector = self.configurators['scale']
            if force or abs(scale_config.value - self.last_rescale) > (scale_config.scale.get_adjustment().get_step_increment() - 0.0001):
                self.last_rescale = scale_config.value
                thumbnail = pil_scale(self.page_image, None, int(scale_config.get_exp() * self.page_image.height))
                self.image.set_from_pixbuf(pil_to_pixbuf(thumbnail))
        else:
            self.image.set_from_icon_name('missing-image', Gtk.IconSize.DIALOG)
        self.update_transformation()

    def _on_mouse(self, _widget: Gtk.Overlay, e: Gdk.EventButton) -> None:
        if self.t is None or self.region_map is None:
            return

        tx, ty = self.t.transform(e.x, e.y)
        if self.region_map.find_region(tx, ty, ignore_regions=[]):
            # noinspection PyArgumentList
            cursor = Gdk.Cursor.new_from_name(self.container.get_display(), 'pointer')
            self.container.get_window().set_cursor(cursor)
            return

        # noinspection PyArgumentList
        cursor = Gdk.Cursor.new_from_name(self.container.get_display(), 'default')
        self.container.get_window().set_cursor(cursor)

    def _on_button(self, _widget: Gtk.Overlay, e: Gdk.EventButton) -> bool:
        _widget.grab_focus()
        old_region = self.current_region
        if self.t is None or self.region_map is None:
            self.current_region = None
        else:
            tx, ty = self.t.transform(e.x, e.y)
            self.current_region = self.region_map.find_region(tx, ty)

        if self.current_region is not old_region:
            self.invalidate_region(self.current_region)
            self.invalidate_region(old_region)
            self.update_status_bar()

        if e.button == Gdk.BUTTON_SECONDARY and e.type == Gdk.EventType.BUTTON_PRESS:
            if self.current_region:
                self._on_context_menu(e, self.current_region)
        return False

    def _on_context_menu(self, e: Gdk.EventButton, r: Region) -> bool:
        p = Gtk.Menu(visible=True)
        p.append(Gtk.MenuItem(visible=True, label=r.id))
        p.append(Gtk.MenuItem(visible=True, label=r.text))
        p.popup_at_pointer(e)
        return False

    def _on_scroll(self, _widget: Gtk.EventBox, event: Gdk.EventScroll) -> bool:
        """
        Handles zoom in / zoom out on Ctrl+mouse wheel
        """
        _widget.grab_focus()
        accel_mask = Gtk.accelerator_get_default_mod_mask()
        if event.state & accel_mask == Gdk.ModifierType.CONTROL_MASK:
            did_scroll, delta_x, delta_y = event.get_scroll_deltas()
            if did_scroll and abs(delta_y) > 0:
                scale_config: ImageZoomSelector = self.configurators['scale']
                scale_config.set_value(self.scale + delta_y * 0.1)
                return True
        return False

    def _on_viewport_size_allocate(self, _sender: Gtk.Widget, rect: Gdk.Rectangle) -> None:
        self.viewport_size = rect
        self.update_transformation()

    def _on_zoom_by(self, _action: Gio.SimpleAction, steps_v: Optional[GLib.Variant] = None) -> None:
        scale_config: ImageZoomSelector = self.configurators['scale']
        scale_config.zoom_by(steps_v.get_int32())
        self.update_transformation()

    def _on_zoom_to(self, _action: Gio.SimpleAction, to_v: Optional[GLib.Variant] = None) -> None:
        scale_config: ImageZoomSelector = self.configurators['scale']
        scale_config.zoom_to(
            to_v.get_string(),
            self.viewport_size.width / self.page_image.width,
            self.viewport_size.height / self.page_image.height
        )
        self.update_transformation()

    def _query_tooltip(self, _image: Gtk.Image, x: int, y: int, _keyboard_mode: bool, tooltip: Gtk.Tooltip) -> bool:
        if self.t is None or self.region_map is None:
            return False

        tx, ty = self.t.transform(x, y)

        region = self.region_map.find_region(tx, ty, ignore_regions=[])

        content = '<tt>{:d}, {:d}</tt>'.format(int(tx), int(ty))
        if region:
            content += '\n<tt><big>{}</big></tt>\n\n{}\n'.format(str(region), escape(region.text))

            if region.text_conf:
                content += '\n<tt>@text.conf=</tt>{}'.format(region.text_conf)

            if region.coords_conf:
                content += '\n<tt>@coords.conf=</tt>{}'.format(region.coords_conf)
            if region.region_subtype:
                content += '\n<tt>@type:</tt> {}'.format(region.region_subtype)
            for attribute in [
                    'custom',
                    'comments',
                    'production',
                    'orientation',
                    'leading',
                    'indented',
                    'continuation',
                    'readingDirection',
                    'readingOrientation',
                    'textLineOrder',
                    'primaryLanguage',
                    'secondaryLanguage',
                    'language',
                    'primaryScript',
                    'secondaryScript',
                    'script',
                    'ligature',
                    'symbol',
                    'colourDepth',
                    'bgColour',
                    'embText',
                    'penColour',
                    'numColours',
                    'lineColour',
                    'lineSeparators',
                    'rows',
                    'columns',
                    'colour',
            ]:
                if hasattr(region.region, attribute) and getattr(region.region, attribute):
                    content += '\n<tt>@{}=</tt>{}'.format(attribute, getattr(region.region, attribute))
            if hasattr(region.region, 'TextStyle') and getattr(region.region, 'TextStyle'):
                style = getattr(region.region, 'TextStyle')
                for attribute in [
                        'textColour',
                        'bgColour',
                        'textColourRgb',
                        'bgColourRgb',
                        'reverseVideo',
                        'fontSize',
                        'xHeight',
                        'kerning',
                        'fontFamily',
                        'serif',
                        'monospace',
                        'bold',
                        'italic',
                        'underlined',
                        'underlineStyle',
                        'subscript',
                        'superscript',
                        'strikethrough',
                        'smallCaps',
                        'letterSpaced',
                ]:
                    if getattr(style, attribute):
                        content += '\n<tt>@{}=</tt>{}'.format(attribute, getattr(style, attribute))
            if region.warnings:
                content += '\n\nWarnings:' + ('\n '.join(region.warnings))

        tooltip.set_markup(content)

        return True

    def update_status_bar(self) -> None:
        for w in self.status_bar.get_children():
            self.status_bar.remove(w)

        if self.current_region:
            for i, r in enumerate(reversed(self.current_region.breadcrumbs())):
                if i:
                    self.status_bar.pack_start(Gtk.Separator(visible=True, orientation=Gtk.Orientation.HORIZONTAL), False, False, 0)

                label = Gtk.Label(visible=True, label=str(r), ellipsize=Pango.EllipsizeMode.MIDDLE, max_width_chars=20, tooltip_text=str(r))
                self.status_bar.pack_start(label, False, False, 5)
        else:
            self.status_bar.pack_start(Gtk.Label(visible=True, label=' '), False, False, 2)

    def invalidate_region(self, r: Region) -> None:
        if self.t is None or r is None:
            return

        x, y, w, h = self.t.transform_region(r.poly.buffer(5))
        if w > 0 and h > 0:
            self.highlight.queue_draw_area(x, y, w, h)

    def draw_highlight(self, _area: Gtk.DrawingArea, context: Context) -> None:
        if self.current_region and self.t:
            poly: Polygon = self.current_region.poly
            poly = poly.buffer(1, single_sided=True)
            # TODO: 239, 134, 97, 0.7 taken from gtk.css, possible to get it from os????
            context.set_source_rgba(239 / 255.0, 134 / 255.0, 97 / 255.0, 0.7)
            context.set_line_width(clamp(self.configurators['scale'].get_exp() * 12, 0.5, 5))
            context.new_path()
            # Nice idea, but didn't work with scrolling: context.set_matrix(Matrix(1.0/self.t.scale, 0, 0, 1.0/self.t.scale, -self.t.tx, -self.t.ty))
            for coord in poly.exterior.coords:
                context.line_to(*self.t.inverse(*coord))
            context.close_path()
            context.stroke()

    def update_transformation(self) -> None:
        if self.page_image is None or self.image is None or self.image.get_pixbuf() is None:
            return

        pb = self.image.get_pixbuf()
        size, _ = self.image.get_allocated_size()

        self.t = Transformation(
            self.page_image.width / pb.get_width(),
            -(size.width - pb.get_width()) * 0.5,
            -size.height * 0.0 + pb.get_height() * 0.0,
            self.page_image.width,
            self.page_image.height
        )
        self.highlight.queue_draw()

    def open_screenshotdialog(self, button: Gtk.Button) -> None:
        if self.page_image is None:
            return

        dialog = Gtk.FileChooserDialog(title="Save image under...",
                                       parent=self.window,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL,
                           Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_SAVE,
                           Gtk.ResponseType.OK)
        filter_png = Gtk.FileFilter()
        filter_png.set_name("PNG image files")
        filter_png.add_mime_type("image/png")
        dialog.add_filter(filter_png)
        dialog.set_current_name("untitled.png")

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
        else:
            filename = ''

        dialog.destroy()
        if filename:
            self.page_image.save(filename)



import unittest
from unittest.mock import MagicMock

from gi.repository import Gtk
from ocrd_browser.model import Document
from ocrd_browser.view import ViewText, ViewEmpty
from ocrd_browser.ui import MainWindow
from ocrd_browser.view.manager import ViewManager
from tests import TestCase


class ManagerTestCase(TestCase):

    def setUp(self):
        self.root = Gtk.Box(name='container')
        self.win = MagicMock(spec=MainWindow)
        self.win.document = Document.create(self.win)
        self.win.current_page_id = None

        self.vm = ViewManager(self.win, self.root)

    def test_can_construct(self):
        self.assertIsNotNone(self.vm)
        self.assertEqual('container\n', self.vm.print_level())

    def test_one_view(self):
        expected = """
container
\tview_0: ViewText
""".strip()
        self.vm.set_root_view(ViewText)
        self.assertEqual(expected, self.vm.print_level().strip())

    def test_two_views(self):
        expected = """
container
\tGtkPaned
\t\tview_0: ViewText
\t\tview_1: ViewEmpty
""".strip()
        self.vm.set_root_view(ViewText)
        self.vm.split('view_0', ViewEmpty)
        self.assertEqual(expected, self.vm.print_level().strip())

    def test_split_and_close(self):
        expected = """
container
\tGtkPaned
\t\tview_0: ViewText
\t\tview_1: ViewEmpty
""".strip()
        self.vm.set_root_view(ViewText)
        v1 = self.vm.split('view_0', ViewEmpty)
        v2 = self.vm.split(v1.name, ViewEmpty)
        v3 = self.vm.split(v2.name, ViewEmpty)
        self.vm.close(v2.name)
        self.vm.close(v3.name)

        self.assertEqual(expected, self.vm.print_level().strip())


if __name__ == '__main__':
    unittest.main()



from gi.repository import Gtk, Gdk, GLib, Gio

from typing import Any, List, Optional, Tuple

from itertools import zip_longest
from ocrd_browser.util.image import pil_to_pixbuf, pil_scale
from ocrd_models.constants import NAMESPACES as NS
from .base import (
    View,
    FileGroupSelector,
    FileGroupFilter,
    PageQtySelector,
    ImageZoomSelector
)
from ..model import Page
from ..util.gtk import WhenIdle, ActionRegistry


class ViewImages(View):
    """
    View of one or more consecutive images
    """

    label = 'Image'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.file_group: Tuple[Optional[str], Optional[str]] = (None, None)
        self.page_qty: int = 1
        self.preview_height: int = 10
        self.scale: float = -2.0
        self.last_rescale = -100
        self.viewport: Optional[Gtk.Viewport] = None
        self.image_box: Optional[Gtk.Box] = None
        self.pages: List[Page] = []

    def build(self) -> None:
        super(ViewImages, self).build()

        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.IMAGE))
        self.add_configurator('page_qty', PageQtySelector())
        self.add_configurator('scale', ImageZoomSelector(2.0, 0.05, -4.0, 2.0))

        self.image_box = Gtk.Box(visible=True, orientation=Gtk.Orientation.HORIZONTAL, homogeneous=True)

        eventbox = Gtk.EventBox(visible=True, can_focus=True, has_focus=True, focus_on_click=True, is_focus=True)
        eventbox.add_events(Gdk.EventMask.SMOOTH_SCROLL_MASK | Gdk.EventMask.BUTTON_PRESS_MASK)
        eventbox.connect('scroll-event', self.on_scroll)
        eventbox.connect('button-press-event', self.on_button)
        eventbox.add(self.image_box)

        self.viewport = Gtk.Viewport(visible=True, hscroll_policy='natural', vscroll_policy='natural')
        self.viewport.connect('size-allocate', self._on_viewport_size_allocate)
        self.viewport.add(eventbox)

        self.scroller.add(self.viewport)

        actions = ActionRegistry()
        actions.create(name='zoom_by', param_type=GLib.VariantType('i'), callback=self._on_zoom_by)
        actions.create(name='zoom_to', param_type=GLib.VariantType('s'), callback=self._on_zoom_to)
        eventbox.insert_action_group("view", actions.for_widget)

        self.rebuild_pages()

    def config_changed(self, name: str, value: Any) -> None:
        super(ViewImages, self).config_changed(name, value)
        if name == 'page_qty':
            WhenIdle.call(self.rebuild_pages, priority=1)
        if name == 'file_group':
            WhenIdle.call(self.reload, priority=10)
        if name == 'scale':
            WhenIdle.call(self.rescale)

    def rebuild_pages(self) -> None:
        existing_pages = {child.get_name(): child for child in self.image_box.get_children()}

        # We need a variable number of Gtk.Image (depending on number of AlternativeImage)
        # in a fixed number of Gtk.VBox (depending on configured page_qty)
        # in a single Gtk.HBox (for the current view).
        # So whenever page_qty changes, some HBoxes will be re-used,
        # and whenever page_id changes, some VBoxes will be re-used.
        for i in range(0, self.page_qty):
            name = 'page_{}'.format(i)
            if not existing_pages.pop(name, None):
                page = Gtk.Box(visible=True, orientation=Gtk.Orientation.VERTICAL, homogeneous=False, spacing=0)
                self.image_box.add(page)

        for child in existing_pages.values():
            child.destroy()

        WhenIdle.call(self.reload, priority=10)

    def page_activated(self, _sender: Gtk.Widget, page_id: str) -> None:
        self.page_id = page_id
        WhenIdle.call(self.reload)

    @property
    def use_file_group(self) -> str:
        return self.file_group[0]

    def reload(self) -> None:
        if self.document:
            display_ids = self.document.display_id_range(self.page_id, self.page_qty)
            self.pages = []
            for display_id in display_ids:
                self.pages.append(self.document.page_for_id(display_id, self.use_file_group))
        WhenIdle.call(self.redraw)

    def redraw(self) -> None:
        if self.pages:
            box: Gtk.Box
            for box, page in zip_longest(self.image_box.get_children(), self.pages):
                existing_images = {child.get_name(): child for child in box.get_children()}
                for i, img in enumerate(page.images if page else [None]):
                    name = 'image_{}'.format(i)
                    image: Gtk.Image
                    image = existing_images.pop(name, None)
                    if not image:
                        image = Gtk.Image(name=name, visible=True,
                                          icon_name='gtk-missing-image',
                                          icon_size=Gtk.IconSize.DIALOG)
                        box.add(image)
                    if img:
                        if page.image_files[0] == page.file:
                            # PAGE-XML was created from the (first) image file directly
                            image.set_tooltip_text(page.id)
                        else:
                            img_file = page.image_files[i]
                            # get segment ID for AlternativeImage as tooltip
                            img_id = page.pc_gts.gds_elementtree_node_.xpath(
                                '//page:AlternativeImage[@filename="{}"]/../@id'.format(img_file.local_filename),
                                namespaces=NS)
                            if img_id:
                                image.set_tooltip_text(page.id + ':' + img_id[0])
                            else:
                                image.set_tooltip_text(img_file.local_filename)
                    else:
                        image.set_from_icon_name('missing-image', Gtk.IconSize.DIALOG)
                for child in existing_images.values():
                    child.destroy()
            WhenIdle.call(self.rescale, force=True)

    def rescale(self, force: bool = False) -> None:
        if self.pages:
            box: Gtk.Box
            scale_config: ImageZoomSelector = self.configurators['scale']
            if force or abs(scale_config.value - self.last_rescale) > (scale_config.scale.get_adjustment().get_step_increment() - 0.0001):
                self.last_rescale = scale_config.value
                for box, page in zip_longest(self.image_box.get_children(), self.pages):
                    images = {child.get_name(): child for child in box.get_children()}
                    for i, img in enumerate(page.images if page else [None]):
                        name = 'image_{}'.format(i)
                        image: Gtk.Image
                        image = images[name]
                        if img:
                            thumbnail = pil_scale(img, None, int(scale_config.get_exp() * img.height))
                            image.set_from_pixbuf(pil_to_pixbuf(thumbnail))

    def on_button(self, _widget: Gtk.EventBox, event: Gdk.EventButton) -> bool:
        _widget.grab_focus()
        return False

    def on_scroll(self, _widget: Gtk.EventBox, event: Gdk.EventScroll) -> bool:
        _widget.grab_focus()
        # Handles zoom in / zoom out on Ctrl+mouse wheel
        accel_mask = Gtk.accelerator_get_default_mod_mask()
        if event.state & accel_mask == Gdk.ModifierType.CONTROL_MASK:
            did_scroll, delta_x, delta_y = event.get_scroll_deltas()
            if did_scroll and abs(delta_y) > 0:
                scale_config: ImageZoomSelector = self.configurators['scale']
                scale_config.set_value(self.scale + delta_y * 0.1)
                return True
        return False

    def _on_viewport_size_allocate(self, _sender: Gtk.Widget, rect: Gdk.Rectangle) -> None:
        self.viewport_size = rect

    def _on_zoom_by(self, _action: Gio.SimpleAction, steps_v: Optional[GLib.Variant] = None) -> None:
        scale_config: ImageZoomSelector = self.configurators['scale']
        scale_config.zoom_by(steps_v.get_int32())

    def _on_zoom_to(self, _action: Gio.SimpleAction, to_v: Optional[GLib.Variant] = None) -> None:
        scale_config: ImageZoomSelector = self.configurators['scale']
        all_width = 0
        all_height = 0
        for page in self.pages:
            page_height = 0
            page_width = 0
            for image in page.images:
                page_height += image.height
                page_width = max(page_width, image.width)
            all_height = max(all_height, page_height)
            all_width += page_width

        scale_config.zoom_to(
            to_v.get_string(),
            self.viewport_size.width / all_width,
            self.viewport_size.height / all_height
        )



from tests import TestCase
from ocrd_browser.model.page_xml_renderer import RegionFactory, Region
from ocrd_models.ocrd_page import CoordsType, SeparatorRegionType


class RegionFactoryTestCase(TestCase):

    def setUp(self) -> None:
        identity = {
            'transform': [[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]],
            'angle': 0,
            'features': ',normalized'
        }
        self.factory: RegionFactory = RegionFactory(identity, 'DUMMY_0001', None)

    def test_create(self):
        ds = SeparatorRegionType(id='r6', Coords=CoordsType(points="0,0 0,1 1,1 1,0"))
        try:
            with self.assertLogs('ocrd_browser.model.page_xml_renderer', level='WARNING') as log_watch:
                region = self.factory.create(ds)
                raise ValueError('Dummy instead of assertNoLogs')
        except ValueError:
            pass
        self.assertEqual(len(log_watch.output), 0, '{:d} Warning(s) logged "{:s}'.format(len(log_watch.output), '\n'.join(log_watch.output)))
        self.assertIsInstance(region, Region)
        self.assertGreater(region.poly.area, 0)

    def test_create_with_error(self):
        ds = SeparatorRegionType(id='r6', Coords=CoordsType(points="1,1 1,1 1,1 1,1"))
        with self.assertLogs('ocrd_browser.model.page_xml_renderer', level='ERROR') as log_watch:
            region = self.factory.create(ds)
        self.assertRegex(log_watch.output[0], r'ERROR:ocrd_browser\.model\.page_xml_renderer\.RegionFactory:Page "DUMMY_0001" @ SeparatorRegion#r6 Too few points.+')
        self.assertIsNone(region)

    def test_create_with_warning(self):
        ds = SeparatorRegionType(id='r6', Coords=CoordsType(points="239,1303 508,1303 899,1302 1626,1307 2441,1307 2444,1319 2414,1322 1664,1319 619,1317 235,1317 237,1302 235,1302"))
        with self.assertLogs('ocrd_browser.model.page_xml_renderer', level='WARNING') as log_watch:
            region = self.factory.create(ds)
        self.assertIsNotNone(region)
        self.assertRegex(log_watch.output[0], r'WARNING:ocrd_browser\.model\.page_xml_renderer\.RegionFactory:Page "DUMMY_0001" @ SeparatorRegion#r6 Self-intersection.+')
        self.assertRegex(region.warnings[0], r'Self-intersection.+')

    def test_create_with_warning_negative(self):
        ds = SeparatorRegionType(id='r6', Coords=CoordsType(points="0,0 0,-1 -1,-1 -1,0"))
        with self.assertLogs('ocrd_browser.model.page_xml_renderer', level='WARNING'):
            region = self.factory.create(ds)
        self.assertRegex(region.warnings[0], r'is negative')

    def test_create_with_warning_too_few_points(self):
        ds = SeparatorRegionType(id='r6', Coords=CoordsType(points="0,0 0,1 1,1"))
        with self.assertLogs('ocrd_browser.model.page_xml_renderer', level='WARNING'):
            region = self.factory.create(ds)
        self.assertRegex(region.warnings[0], r'has too few points')



import shlex
from subprocess import Popen
from typing import Dict, Optional, Any

from ocrd_models import OcrdFile
from ocrd_utils import getLogger

from ocrd_browser.model import Document
from ocrd_browser.util.config import _Tool, Settings


class ResolvableFileName:
    def __init__(self, filename: str, in_doc: Document):
        self.filename = filename
        self.in_doc = in_doc

    @property
    def absolute(self) -> str:
        return shlex.quote(str(self.in_doc.path(self.filename).absolute()))

    @property
    def relative(self) -> str:
        return shlex.quote(self.filename)

    def __str__(self) -> str:
        return self.absolute


class QuotingProxy:
    def __init__(self, object_: Any):
        self.object = object_

    def __getattr__(self, item: str) -> str:
        if hasattr(self.object, item):
            return shlex.quote(getattr(self.object, item))
        else:
            raise AttributeError("Unknown attribute '{}' on {} ".format(item, self.object))


class FileProxy(QuotingProxy):
    def __init__(self, file: OcrdFile, in_doc: Document):
        super().__init__(file)
        self.in_doc = in_doc

    @property
    def local_filename(self) -> ResolvableFileName:
        return ResolvableFileName(self.object.local_filename, self.in_doc)

    # Recommended alias path, because local_filename sounds like a relative path
    path = local_filename


class Launcher:
    def __init__(self, tools: Optional[Dict[str, _Tool]] = None):
        self.tools = tools if tools else Settings.get().tools

    def launch(self, toolname: str, doc: Document, file: OcrdFile) -> Optional[Popen]:  # type: ignore[type-arg]
        if toolname in self.tools:
            return self.launch_tool(self.tools[toolname], doc, file)
        else:
            log = getLogger('ocrd_browser.util.launcher.Launcher.launch')
            log.error(
                'Tool "%s" not found in your config, to fix place the following section in your ocrd-browser.conf',
                toolname)
            log.error('[Tool %s]', toolname)
            log.error('commandline = /usr/bin/yourtool --base-dir {workspace.directory} {file.path.absolute}')
            return None

    def launch_tool(self, tool: _Tool, doc: Document, file: OcrdFile, **kwargs: Any) -> Popen:  # type: ignore[type-arg]
        log = getLogger('ocrd_browser.util.launcher.Launcher.launch_tool')
        commandline = self._template(tool.commandline, doc, file)
        log.debug('Calling tool "%s" with commandline: ', tool.name)
        log.debug('%s', commandline)
        process = Popen(args=commandline, shell=True, cwd=doc.directory, **kwargs)
        return process

    def _template(self, arg: str, doc: Document, file: OcrdFile) -> str:
        return arg.format(file=FileProxy(file, doc), workspace=QuotingProxy(doc.workspace))



import unittest
from ocrd_browser.view import ViewImages, ViewRegistry
from tests import TestCase


class ViewManagerTestCase(TestCase):

    def setUp(self):
        self.vm = ViewRegistry({'images': (ViewImages, 'Images', 'Displays Images')})

    def test_get_view_options(self):
        expected = {'images': 'Images'}
        actual = self.vm.get_view_options()
        self.assertEqual(expected, actual)

    def test_get_view(self):
        expected = ViewImages
        actual = self.vm.get_view('images')
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



from gi.repository import Gtk, GLib

from typing import Callable, Sequence, Dict, Optional, Any, Iterator
from concurrent.futures import Future, ThreadPoolExecutor, as_completed

RowInitCallback = Callable[[Gtk.TreeModelRow], None]
RowLoadCallback = Callable[[Gtk.TreeModelRow], Gtk.TreeModelRow]
RowHashCallback = Callable[[Gtk.TreeModelRow], str]


class LazyLoadingListStore(Gtk.ListStore):

    def __init__(self, *column_types: type, init_row: RowInitCallback, load_row: RowLoadCallback,
                 hash_row: RowHashCallback):
        column_type_list = list(column_types)
        column_type_list.append(str)
        super().__init__(*column_type_list)
        self.init_row = init_row
        self.load_row = load_row
        self.hash_row = hash_row
        self.futures: Optional[Dict[Future[Gtk.TreeModelRow], Gtk.TreeModelRow]] = None
        self.row_inserted_handler = self.connect('row-inserted', self._on_row_inserted)
        self.row_changed_handler = self.connect('row-changed', self._on_row_changed)

    def start_loading(self) -> None:
        self.futures = {}
        self.submit_all()
        GLib.timeout_add(10, self._collect_workers().__next__, priority=GLib.PRIORITY_LOW)

    def submit_all(self) -> bool:
        pool = ThreadPoolExecutor()
        for row in self:
            self._submit_future(row, pool)
        pool.shutdown(wait=False)
        return True

    def _submit_future(self, row: Gtk.TreeModelRow, pool: Optional[ThreadPoolExecutor] = None) -> None:
        if self.futures is None:
            return
        own_pool = pool is None
        if own_pool:
            pool = ThreadPoolExecutor()

        row_hash = self.hash_row(row)

        if row[-1] != row_hash and row not in self.futures.values():
            future = pool.submit(self.load_row, row[:])
            self.futures[future] = row

        if own_pool:
            pool.shutdown(wait=False)

    def _do_insert(self, position: int, row: Sequence[Any]) -> None:
        if row is not None:
            row = tuple(row) + (None,)
        super()._do_insert(position, row)

    def _on_row_inserted(self, list_store: Gtk.ListStore, _path: Gtk.TreePath, it: Gtk.TreeIter) -> None:
        row = list_store[it]
        with self.handler_block(self.row_changed_handler):
            self.init_row(row)
        self._submit_future(row)

    def _on_row_changed(self, list_store: Gtk.ListStore, _path: Gtk.TreePath, it: Gtk.TreeIter) -> None:
        row = list_store[it]
        self._submit_future(row)

    def _collect_workers(self) -> Iterator[bool]:
        for future in as_completed(self.futures):
            row = self.futures.pop(future)
            try:
                new_row_data = future.result()
            except Exception as exc:
                print('{} generated an exception: {}'.format(row[0], exc))
            else:
                with self.handler_block(self.row_changed_handler):
                    for i, (old, new) in enumerate(zip(row[:], new_row_data)):
                        if old != new:
                            row[i] = new_row_data[i]
                    row[-1] = self.hash_row(row)
                yield True
        # Futures are finished for now, check back every 50ms if there is something new
        GLib.timeout_add(50, self._collect_workers().__next__, priority=GLib.PRIORITY_LOW)
        yield False



from __future__ import annotations
from typing import Optional, Tuple, List, Set, Union, cast, Callable, Any, Dict, TYPE_CHECKING

import atexit
import errno
import os
import shutil
from functools import wraps

from ocrd import Resolver
from ocrd_browser.model.page import Page
from ocrd_browser.util.file_groups import best_file_group
from ocrd_browser.util.image import add_dpi_to_png_buffer
from ocrd_browser.util.streams import SilencedStreams
from ocrd_modelfactory import page_from_file
from ocrd_models.constants import NAMESPACES as NS
from ocrd_models import OcrdFile
from ocrd_utils import pushd_popd
from ocrd_utils.constants import MIME_TO_EXT
from ocrd_utils import getLogger

from pathlib import Path
from tempfile import mkdtemp
from datetime import datetime
from urllib.parse import urlparse, unquote
from PIL import Image

import cv2

if TYPE_CHECKING:
    from ocrd import Workspace
    from ocrd_models.ocrd_page_generateds import PcGtsType
    # noinspection PyProtectedMember
    from lxml.etree import ElementBase as Element, _ElementTree as ElementTree
    from numpy import array as ndarray

EventCallBack = Optional[Callable[[str, Any], None]]


def check_editable(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def guard(self: 'Document', *args: List[Any], **kwargs: Dict[Any, Any]) -> Any:
        if self._editable is not True:
            raise PermissionError('Document is not editable, can not call  {}'.format(func.__qualname__))
        return_value = func(self, *args, **kwargs)
        return return_value

    return guard


class Document:
    temporary_workspaces: List[str] = []

    def __init__(self, workspace: Optional[Workspace], emitter: Optional[EventCallBack] = None, editable: bool = False,
                 original_url: str = None):
        self.workspace = workspace
        self.emitter = emitter
        self._original_url = original_url
        self._editable = editable
        self._empty = True
        self._modified = False
        if self.workspace:
            os.chdir(self.workspace.directory)

    @classmethod
    def create(cls, emitter: EventCallBack = None) -> Document:
        return cls(None, emitter=emitter)

    @classmethod
    def load(cls, mets_url: Union[Path, str] = None, emitter: EventCallBack = None) -> Document:
        """
        Load a project from an url as a readonly view

        If you want to modify the Workspace, use Document.clone instead
        """
        if not mets_url:
            return cls.create(emitter=emitter)
        mets_path = cls._to_path(mets_url)

        workspace = Resolver().workspace_from_url(str(mets_path), download=False)
        doc = cls(workspace, emitter=emitter, original_url=str(mets_url))
        doc._empty = False
        return doc

    @classmethod
    def clone(cls, mets_url: Union[Path, str], emitter: EventCallBack = None, editable: bool = True) -> Document:
        """
        Clones a project (mets.xml and all used files) to a temporary directory for editing
        """
        doc = cls(cls._clone_workspace(mets_url), emitter=emitter, editable=editable, original_url=str(mets_url))
        doc._empty = False
        return doc

    @classmethod
    def _clone_workspace(cls, mets_url: Union[Path, str]) -> Workspace:
        """
        Clones a workspace (mets.xml and all used files) to a temporary directory for editing
        """
        log = getLogger('ocrd_browser.model.document.Document._clone_workspace')
        mets_url = cls._strip_local(mets_url, disallow_remote=False)
        temporary_workspace = mkdtemp(prefix='browse-ocrd-clone-')
        cls.temporary_workspaces.append(temporary_workspace)
        # TODO download = False and lazy loading would be nice for responsiveness
        log.info("Cloning '%s' to '%s'", mets_url, temporary_workspace)
        workspace = Resolver().workspace_from_url(mets_url=mets_url, dst_dir=temporary_workspace, download=True)
        return workspace

    @check_editable
    def save(self, backup_directory: Union[bool, Path, str] = True) -> None:
        if not self._original_url:
            raise ValueError('Need an _original_url to save')
        self.save_as(self._original_url, backup_directory=backup_directory)

    def save_as(self, mets_url: Union[Path, str], backup_directory: Union[bool, Path, str] = True) -> None:
        log = getLogger('ocrd_browser.model.document.Document.save_as')
        mets_path = self._to_path(mets_url)

        workspace_directory = mets_path.parent
        if workspace_directory.exists():
            if backup_directory:
                if isinstance(backup_directory, bool):
                    backup_directory = self._derive_backup_directory(workspace_directory)
                shutil.move(str(workspace_directory), str(backup_directory))
            else:
                shutil.rmtree(str(workspace_directory))

        mets_basename = mets_path.name
        workspace_directory.mkdir(parents=True, exist_ok=True)
        self._emit('document_saving', 0, None)

        saved_space = Resolver().workspace_from_url(mets_url=self.workspace.mets_target, mets_basename=mets_basename,
                                                    download=False, clobber_mets=True, dst_dir=workspace_directory)
        saved_files = list(saved_space.mets.find_files())
        for n, f in enumerate(saved_files):
            f = saved_space.download_file(f)
            self._emit('document_saving', n / len(saved_files), f)

        self._emit('document_saving', 1, None)
        self._emit('document_saved', Document(saved_space, self.emitter))
        self._original_url = str(mets_path)
        self._modified = False
        log.info('Saved to %s', self._original_url)

    @property
    def directory(self) -> Path:
        """
        Get workspace directory as a Path object
        """
        return Path(self.workspace.directory) if self.workspace else None

    @property
    def mets_filename(self) -> str:
        """
        Gets the mets file name (e.g. "mets.xml")
        """
        return Path(self.workspace.mets_target).name if self.workspace else 'mets.xml'

    @property
    def baseurl_mets(self) -> str:
        """
        Gets the uri of the original mets file name
        """
        return str(self.workspace.baseurl) + '/' + self.mets_filename if self.workspace else None

    def path(self, other: Union[OcrdFile, Path, str]) -> Optional[Path]:
        """
        Resolves other relative to current workspace
        """
        if not self.directory:
            return None
        if isinstance(other, OcrdFile):
            return self.directory.joinpath(other.local_filename)
        elif isinstance(other, Path):
            return self.directory.joinpath(other)
        elif isinstance(other, str):
            return self.directory.joinpath(other)
        else:
            raise ValueError('Unsupported other of type {}'.format(type(other)))

    @property
    def _tree(self) -> Optional[ElementTree]:
        # noinspection PyProtectedMember
        return self.workspace.mets._tree if self.workspace else None

    def xpath(self, xpath: str) -> Any:
        return self._tree.getroot().xpath(xpath, namespaces=NS) if self.workspace else []

    @property
    def page_ids(self) -> List[str]:
        """
        List of page_ids in this workspace

        TODO: This is unsorted (or sorted in document order) use @ORDER attribute or keep document order in sync with the correct order

        @return: List[str]
        """
        # noinspection PyTypeChecker
        return cast(List[str], self.workspace.mets.physical_pages if self.workspace else [])

    @property
    def file_groups(self) -> List[str]:
        """
        List of file_groups in this workspace

        @return: List[str]
        """
        # noinspection PyTypeChecker
        return cast(List[str], self.workspace.mets.file_groups) if self.workspace else []

    # noinspection PyProtectedMember
    @property
    def mime_types(self) -> Set[str]:
        """
        Set with the distinct mime-types in this workspace

        @return: Set[str]
        """
        return {el.get('MIMETYPE') for el in
                self.xpath('mets:fileSec/mets:fileGrp/mets:file[@MIMETYPE]')}

    @property
    def file_groups_and_mimetypes(self) -> List[Tuple[str, str]]:
        """
        A list with the distinct file_group/mimetype pairs in this workspace

        @return: List[Tuple[str,str]]
        """
        distinct_groups: Dict[Tuple[str, str], None] = {}
        for el in self.xpath('mets:fileSec/mets:fileGrp[@USE]/mets:file[@MIMETYPE]'):
            distinct_groups[(el.getparent().get('USE'), el.get('MIMETYPE'))] = None

        return list(distinct_groups.keys())

    @property
    def title(self) -> str:
        return str(self.workspace.mets.unique_identifier) if self.workspace and self.workspace.mets.unique_identifier else '<unnamed>'

    def get_file_index(self) -> Dict[str, OcrdFile]:
        """
        Return all OcrdFiles by file id and additionally augments the OcrdFile with static_page_id for fast(er) lookup

        Example:
        page17 = [file for file in file_index.values() if file.static_page_id == 'PHYS_0017']

        """
        log = getLogger('ocrd_browser.model.document.Document.get_file_index')
        file_index = {}
        if self.workspace:
            for file in self.workspace.mets.find_files():
                file.static_page_id = None
                file_index[file.ID] = file

        file_pointers: List[Element] = self.xpath(
            'mets:structMap[@TYPE="PHYSICAL"]/mets:div[@TYPE="physSequence"]/mets:div[@TYPE="page"]/mets:fptr')
        for file_pointer in file_pointers:
            file_id = file_pointer.get('FILEID')
            page_id = file_pointer.getparent().get('ID')
            if file_id in file_index:
                file_index[file_id].static_page_id = page_id
            else:
                log.warning("FILEID '%s' for PAGE '%s' not in mets:fileSec", file_id, page_id)

        return file_index

    def get_image_paths(self, file_group: str) -> Dict[str, Path]:
        """
        Builds a Dict ID->Path for all page_ids fast

        More precisely:  fast = Faster than iterating over page_ids and using mets.get_physical_page_for_file for each entry
        """
        log = getLogger('ocrd_browser.model.document.Document.get_image_paths')
        image_paths = {}
        file_index = self.get_file_index()
        for page_id in self.page_ids:
            images = [image for image in file_index.values() if
                      image.static_page_id == page_id and image.fileGrp == file_group]
            if len(images) == 1:
                image_paths[page_id] = self.path(images[0])
            else:
                log.warning('Found %d images for PAGE %s and fileGrp %s, expected 1', len(images), page_id, file_group)
                image_paths[page_id] = None
        return image_paths

    def get_default_image_group(self, preferred_image_file_groups: Optional[List[str]] = None) -> Optional[str]:
        best_group = best_file_group(self.file_groups_and_mimetypes, preferred_image_file_groups, [r'image/.*'], cutoff=0)
        return best_group[0] if best_group else None

    def get_unused_page_id(self, template_page_id: str = 'PAGE_{page_nr}') -> Tuple[str, int]:
        """
        Finds a page_nr that yields an unused page_id for the workspace and returns page_id, page_nr

        @param template_page_id: str
        @return: Tuple[str, int]
        """
        page_nr = len(self.page_ids) + 1
        page_ids = self.page_ids
        while page_nr < 9999:
            page_id = template_page_id.format(**{'page_nr': page_nr})
            if page_id not in page_ids:
                return page_id, page_nr
            page_nr += 1

        raise RuntimeError('No unused page_id found')

    def display_id_range(self, page_id: str, page_qty: int) -> List[str]:
        """
        Calculates an page_id range of size page_qty around page_id
        @param page_id:
        @param page_qty:
        @return:
        """
        if not page_id:
            return []
        try:
            index = self.page_ids.index(page_id)
        except ValueError:
            return []
        index = index - index % page_qty
        return self.page_ids[index:index + page_qty]

    def page_for_id(self, page_id: str, file_group: str) -> Optional['Page']:
        log = getLogger('ocrd_browser.model.document.Document.page_for_id')
        if not page_id:
            return None
        page = Page(self, page_id, file_group)
        if not page.file:
            log.warning("No PAGE-XML and no image for page '{}' in fileGrp '{}'".format(page_id, file_group))
            return None

        return page

    def files_for_page_id(self, page_id: str, file_group: str = None, mimetype: str = None) -> List[OcrdFile]:
        with pushd_popd(self.workspace.directory):
            files: List[OcrdFile] = self.workspace.mets.find_files(fileGrp=file_group, pageId=page_id,
                                                                   mimetype=mimetype)
            files = [self.workspace.download_file(file) for file in files]
            return files

    def page_for_file(self, page_file: OcrdFile) -> PcGtsType:
        # cd and silence Warning: Value "ocrd-cis-word-alignment" ... does not match xsd enumeration restriction on TextDataTypeSimpleType
        with pushd_popd(self.workspace.directory), SilencedStreams(False, True):
            return page_from_file(page_file)

    def resolve_image(self, image_file: OcrdFile) -> Image:
        with pushd_popd(self.workspace.directory):
            pil_image = Image.open(self.workspace.download_file(image_file).local_filename)
            # pil_image.load()
            return pil_image

    @check_editable
    def reorder(self, ordered_page_ids: List[str]) -> None:
        """
        Orders the pages in physSequence according to ordered_page_ids

        """
        log = getLogger('ocrd_browser.model.document.Document.reorder')
        old_page_ids = self.page_ids

        if set(old_page_ids) != set(ordered_page_ids):
            raise ValueError('page_ids do not match: missing in mets.xml: {} / missing in order: {}'.format(
                set(ordered_page_ids).difference(set(old_page_ids)),
                set(self.page_ids).difference(set(ordered_page_ids))
            ))
        log.info('Reordering %s to %s', old_page_ids, ordered_page_ids)
        page_sequence: Element = self.xpath('mets:structMap[@TYPE="PHYSICAL"]/mets:div[@TYPE="physSequence"]')[0]

        ordered_divs = []
        for page_id in ordered_page_ids:
            divs = page_sequence.xpath('mets:div[@TYPE="page"][@ID="%s"]' % page_id, namespaces=NS)
            if divs:
                ordered_divs.append(divs[0])
                page_sequence.remove(divs[0])

        if len(page_sequence) > 0:
            raise RuntimeError('page_sequence not empty, still has: {}'.format(page_sequence.getchildren()))

        for div in ordered_divs:
            page_sequence.append(div)

        old_to_new = dict(zip(old_page_ids, self.page_ids))
        self.save_mets()
        self._emit('document_changed', 'reordered', old_to_new)

    @check_editable
    def delete_images(self, page_id: str, file_group: str = 'OCR-D-IMG') -> List[OcrdFile]:
        image_files: List[OcrdFile] = list(self.workspace.mets.find_files(pageId=page_id, fileGrp=file_group,
                                                                          local_only=True, mimetype='//image/.+'))

        for image_file in image_files:
            self.workspace.remove_file(image_file, force=True, keep_file=False, page_recursive=True,
                                       page_same_group=True)
        self.save_mets()
        self._emit('document_changed', 'page_changed', [page_id])
        return image_files

    def save_mets(self) -> None:
        if not self._editable:
            raise PermissionError('Can not modify Document with _editable == False')
        self.workspace.save_mets()
        self._modified = True

    @check_editable
    def delete_page(self, page_id: str) -> None:
        files = self.workspace.mets.find_files(pageId=page_id, local_only=True)
        for file in files:
            self.workspace.remove_file(file, force=False, keep_file=False)
        self.workspace.mets.remove_physical_page(page_id)
        self.save_mets()
        self._emit('document_changed', 'page_deleted', [page_id])

    @check_editable
    def add_image(self, image: ndarray, page_id: str, file_id: str, file_group: str = 'OCR-D-IMG', dpi: int = 300,
                  mimetype: str = 'image/png') -> OcrdFile:
        extension = MIME_TO_EXT[mimetype]
        retval, image_array = cv2.imencode(extension, image)
        image_bytes = add_dpi_to_png_buffer(image_array.tostring(), dpi)
        local_filename = Path(file_group, '%s%s' % (file_id, extension))
        current_file = self.workspace.add_file(file_group, ID=file_id, mimetype=mimetype, force=True,
                                               content=image_bytes,
                                               local_filename=str(local_filename), pageId=page_id)
        self._empty = False
        self.save_mets()
        self._emit('document_changed', 'page_added', [page_id])
        return current_file

    @property
    def modified(self) -> bool:
        return self._modified

    @property
    def empty(self) -> bool:
        return self._empty

    @property
    def original_url(self) -> str:
        return self._original_url

    @property
    def editable(self) -> bool:
        return self._editable

    @editable.setter
    def editable(self, editable: bool) -> None:
        if editable:
            if self._original_url:
                self.workspace = self._clone_workspace(self._original_url)
            else:
                # noinspection PyTypeChecker
                self.workspace = Resolver().workspace_from_nothing(directory=None, mets_basename='mets.xml')
        else:
            self.workspace = Resolver().workspace_from_url(self.baseurl_mets)
        self._editable = editable
        # self._empty = False
        # self._modified = False

    def _emit(self, event: str, *args: Any) -> None:
        if self.emitter is not None:
            self.emitter(event, *args)

    @staticmethod
    def _strip_local(mets_url: Union[Path, str], disallow_remote: bool = True) -> str:
        result = urlparse(str(mets_url))
        if result.scheme == 'file' or result.scheme == '':
            mets_url = unquote(result.path)
        elif disallow_remote:
            raise ValueError('invalid url {}'.format(mets_url))
        return str(mets_url)

    @staticmethod
    def _to_path(mets_url: Union[Path, str]) -> Path:
        result = urlparse(str(mets_url))
        if not (result.scheme == 'file' or result.scheme == ''):
            raise ValueError('invalid local path/url {}'.format(mets_url))
        return Path(unquote(result.path))

    @staticmethod
    def _derive_backup_directory(workspace_directory: Path, now: datetime = None) -> Path:
        now = now or datetime.now()
        return workspace_directory.parent / ('.bak.' + workspace_directory.name + '.' + now.strftime('%Y%m%d-%H%M%S'))

    @classmethod
    def delete_temporary_workspaces(cls) -> None:
        for temporary_workspace in cls.temporary_workspaces:
            try:
                shutil.rmtree(temporary_workspace)
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise


atexit.register(Document.delete_temporary_workspaces)



from __future__ import annotations
import os
from configparser import ConfigParser
import shlex
from typing import List, Optional, MutableMapping
from shutil import which

from gi.repository import GLib
from ocrd_utils import getLogger

ConfigDict = MutableMapping[str, str]


class _SubSettings:
    @classmethod
    def from_section(cls, _name: str, section: ConfigDict) -> _SubSettings:
        raise NotImplementedError('please override from_section')

    def validate(self) -> None:
        pass

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, repr(vars(self)))


class _FileGroups(_SubSettings):
    def __init__(self, preferred_images: List[str]):
        self.preferred_images = preferred_images

    @classmethod
    def from_section(cls, _name: str, section: ConfigDict) -> _FileGroups:
        preferred_images = section.get('preferredImages', 'OCR-D-IMG, OCR-D-IMG.*')
        return cls(
            [grp.strip() for grp in preferred_images.split(',')]
        )


class _Tool(_SubSettings):
    PREFIX = 'Tool '

    def __init__(self, name: str, commandline: str, shortcut: Optional[str] = None):
        self.name = name
        self.commandline = commandline
        self.shortcut = shortcut

    def validate(self) -> None:
        executable = shlex.split(self.commandline)[0]
        resolved = which(executable)
        if not resolved:
            raise ValueError('Could not locate executable "{}"'.format(executable))

    @classmethod
    def from_section(cls, name: str, section: ConfigDict) -> _Tool:
        return cls(
            name[len(cls.PREFIX):],
            section['commandline'],
            section.get('shortcut', None)
        )


class Settings:

    _settings = None

    def __init__(self, config: ConfigParser, validate: bool = True):
        self.file_groups = _FileGroups.from_section('FileGroups', config['FileGroups'] if 'FileGroups' in config else {'': ''})
        if validate:
            self.file_groups.validate()
        self.tools = {}
        for name, section in config.items():
            if name.startswith(_Tool.PREFIX):
                tool = _Tool.from_section(name, section)
                if validate:
                    tool.validate()
                self.tools[tool.name] = tool

    def __repr__(self) -> str:
        return '{}({})'.format(self.__class__.__name__, repr(vars(self)))

    @classmethod
    def get(cls) -> Settings:
        if cls._settings is None:
            cls._settings = Settings.build_default()
        return cls._settings

    @classmethod
    def build_default(cls, config_dirs: Optional[List[str]] = None, validate: bool = True) -> Settings:
        if config_dirs is None:
            config_dirs = GLib.get_system_config_dirs() + [GLib.get_user_config_dir()]
            try:
                config_dirs.append(os.getcwd())
            except FileNotFoundError:
                # ignore deleted cwd in unit tests
                pass

        config_files = [dir_ + '/ocrd-browser.conf' for dir_ in config_dirs]
        return cls.build_from_files(config_files, validate)

    @classmethod
    def build_from_files(cls, files: List[str], validate: bool = True) -> Settings:
        log = getLogger('ocrd_browser.util.config._Settings.build_from_files')
        config = ConfigParser()
        setattr(config, 'optionxform', lambda option: option)
        read_files = config.read(files)
        log.info('Read %d config file(s): %s, tried %s', len(read_files), ', '.join(read_files), ', '.join(str(file) for file in files))
        return cls(config, validate)



from pathlib import Path
from tempfile import TemporaryDirectory

from tests import TestCase, ASSETS_PATH, TEST_BASE_PATH
from ocrd_browser.model import Document, Page
from datetime import datetime
from ocrd_models.ocrd_page import PcGtsType


# TODO: Later: from tests.assets import Assets, copy_of_directory

class DocumentTestCase(TestCase):

    def setUp(self):
        self.path = ASSETS_PATH / 'kant_aufklaerung_1784/data/mets.xml'

    def test_get_page_ids(self):
        doc = Document.load(self.path)
        self.assertEqual(['PHYS_0017', 'PHYS_0020'], doc.page_ids)

    def test_get_mime_types(self):
        doc = Document.load(self.path)
        self.assertEqual({'application/vnd.prima.page+xml', 'image/tiff', 'application/alto+xml'}, doc.mime_types)

    def test_get_file_groups_and_mimetypes(self):
        doc = Document.load(self.path)
        expected = [
            ('OCR-D-IMG', 'image/tiff'),
            ('OCR-D-GT-PAGE', 'application/vnd.prima.page+xml'),
            ('OCR-D-GT-ALTO', 'application/alto+xml')
        ]
        self.assertEqual(expected, doc.file_groups_and_mimetypes)

    def test_get_page_index(self):
        doc = Document.load(self.path)
        file_index = doc.get_file_index()
        page17 = [file for file in file_index.values() if file.static_page_id == 'PHYS_0017']
        alto = [file for file in file_index.values() if file.mimetype == 'application/alto+xml']
        self.assertEqual(3, len(page17))
        self.assertEqual(2, len(alto))

    def test_get_image_paths(self):
        doc = Document.load(self.path)
        image_paths = doc.get_image_paths('OCR-D-IMG')
        self.assertEqual(2, len(image_paths))
        self.assertEqual('INPUT_0017.tif', image_paths['PHYS_0017'].name)
        self.assertEqual('INPUT_0020.tif', image_paths['PHYS_0020'].name)

    def test_get_default_image_group(self):
        doc = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784-complex/data/mets.xml')
        file_group = doc.get_default_image_group(['OCR-D-IMG-BIN', 'OCR-D-IMG.*'])
        self.assertEqual('OCR-D-IMG-BIN', file_group)

    def test_get_default_image_group_no_preference(self):
        doc = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784-complex/data/mets.xml')
        file_group = doc.get_default_image_group()
        self.assertEqual('OCR-D-IMG', file_group)

    def test_get_default_image_group_with_missing_ocr_d_img(self):
        doc = Document.load(ASSETS_PATH / '../example/workspaces/no_ocrd_d_img_group/mets.xml')
        file_group = doc.get_default_image_group()
        self.assertEqual('OCR-D-IMG-PNG', file_group)

    def test_path_string(self):
        doc = Document.load(self.path)
        self.assertEqual(ASSETS_PATH / 'kant_aufklaerung_1784/data/lala.xml', doc.path('lala.xml'))

    def test_path_path(self):
        doc = Document.load(self.path)
        self.assertEqual(ASSETS_PATH / 'kant_aufklaerung_1784/data/OCR-D-DIR/lala.xml',
                         doc.path(Path('OCR-D-DIR/lala.xml')))

    def test_path_ocrd_file(self):
        doc = Document.load(self.path)
        image_file = list(doc.workspace.mets.find_files(pageId='PHYS_0017', fileGrp='OCR-D-IMG'))[0]
        self.assertEqual(ASSETS_PATH / 'kant_aufklaerung_1784/data/OCR-D-IMG/INPUT_0017.tif', doc.path(image_file))

    def test_reorder(self):
        doc = Document.clone(self.path)
        doc.reorder(['PHYS_0020', 'PHYS_0017'])
        self.assertEqual(['PHYS_0020', 'PHYS_0017'], doc.page_ids)

    def test_reorder_with_wrong_ids_raises_value_error(self):
        doc = Document.clone(self.path)
        with self.assertRaises(ValueError) as context:
            doc.reorder(['PHYS_0021', 'PHYS_0017'])

        self.assertIn('page_ids do not match', str(context.exception))

    def test_delete(self):
        doc = Document.clone(self.path)
        doc.delete_page('PHYS_0017')
        self.assertEqual(['PHYS_0020'], doc.page_ids)

    def test_clone(self):
        doc = Document.clone(self.path)
        self.assertIn('browse-ocrd-clone-', doc.workspace.directory)
        self.assertEqual(str(self.path), doc.baseurl_mets)

        original_files = self.path.parent.rglob('*.*')
        cloned_files = Path(doc.workspace.directory).rglob('*.*')
        for original, cloned in zip(sorted(original_files), sorted(cloned_files)):
            self.assertEqual(original.read_bytes(), cloned.read_bytes())

    def test_save(self):
        doc = Document.clone(self.path)
        with TemporaryDirectory(prefix='browse-ocrd tests') as directory:
            saved_mets = directory + '/mets.xml'
            doc.save_as(saved_mets)
            saved = Document.load(saved_mets)
            self.assertEqual(doc.file_groups, saved.file_groups)
            self.assertEqual(doc.page_ids, saved.page_ids)
            self.assertEqual(doc.workspace.mets.unique_identifier, saved.workspace.mets.unique_identifier)

            for page_id in doc.page_ids:
                for file_group, mime in doc.file_groups_and_mimetypes:
                    original_file = doc.files_for_page_id(page_id, file_group, mime)[0]
                    saved_file = saved.files_for_page_id(page_id, file_group, mime)[0]
                    self.assertEqual(original_file, saved_file)

    def test_derive_backup_directory(self):
        self.assertEqual(
            Path('/home/jk/.bak.important_project.20200813-184321'),
            Document._derive_backup_directory(Path('/home/jk/important_project'), datetime(2020, 8, 13, 18, 43, 21))
        )

    def test_page_for_id_with_no_images_for_page_and_fileGrp(self):
        """
        Issue #4: list index out of range on non-XML fileGrp

        https://github.com/hnesk/browse-ocrd/issues/4
        """
        doc = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784-complex/data/mets.xml')
        with self.assertLogs('ocrd_browser.model.document', level='WARNING') as log_watch:
            page = doc.page_for_id('PHYS_0020', 'OCR-D-IMG-CLIP')
        self.assertIsNone(page)
        self.assertEqual(1, len(log_watch.records))
        self.assertEqual("No PAGE-XML and no image for page 'PHYS_0020' in fileGrp 'OCR-D-IMG-CLIP'",
                         log_watch.records[0].msg)

    def test_page_for_id_with_nothing_for_page_and_fileGrp(self):
        """
        Issue #4 again: This time for missing PAGE-XMLs

        https://github.com/hnesk/browse-ocrd/issues/4
        """
        doc = Document.load(ASSETS_PATH / '../example/workspaces/kant_aufklaerung_1784_missing_xml/mets.xml')
        with self.assertLogs('ocrd_browser.model.document', level='WARNING') as log_watch:
            page = doc.page_for_id('PHYS_0020', 'OCR-D-GT-PAGE')
        self.assertIsNone(page)
        self.assertEqual(1, len(log_watch.records))
        self.assertEqual("No PAGE-XML and no image for page 'PHYS_0020' in fileGrp 'OCR-D-GT-PAGE'",
                         log_watch.records[0].msg)

    def test_page_for_id_with_multiple_images_for_page_and_fileGrp(self):
        """
        returns first image and warns
        """
        doc = Document.load(ASSETS_PATH / 'kant_aufklaerung_1784-complex/data/mets.xml')
        # with self.assertLogs('ocrd_browser.model.document', level='WARNING') as log_watch:
        page = doc.page_for_id('PHYS_0017', 'OCR-D-IMG-CLIP')
        self.assertIsInstance(page, Page)
        self.assertIsInstance(page.pc_gts, PcGtsType)
        # self.assertEqual(1, len(log_watch.records))
        # self.assertEqual("No PAGE-XML but 2 images for page 'PHYS_0017' in fileGrp 'OCR-D-IMG-CLIP'", log_watch.records[0].msg)

    def test_modify_when_not_editable(self):
        doc = Document.load(self.path)
        with self.assertRaises(PermissionError):
            doc.reorder(['PHYS_0020', 'PHYS_0017'])

    def test_modify_when_editable(self):
        doc = Document.clone(self.path)
        doc.reorder(['PHYS_0020', 'PHYS_0017'])

    def test_path_with_spaces(self):
        doc = Document.load((TEST_BASE_PATH / 'example/workspaces/heavy quoting/mets.xml').as_uri())
        page = doc.page_for_id('PHYS_0017', 'OCR-D-GT-PAGE')
        image = doc.workspace.image_from_page(page.page, 'PHYS_0017')
        # Assert no exceptions happened and a sensible return value
        self.assertGreater(image[0].height, 100)

    def test_missing_image(self):
        path = TEST_BASE_PATH / 'example/workspaces/kant_aufklaerung_1784_missing_image/mets.xml'
        uri = path.as_uri()
        doc = Document.load(uri)
        page = doc.page_for_id('PHYS_0017', 'OCR-D-GT-PAGE')
        image, info, exif = page.get_image(feature_selector='', feature_filter='binarized')
        # Assert no exceptions happened and no image returned
        self.assertIsNone(image)



import shlex
import subprocess
import unittest
from ocrd_utils import setOverrideLogLevel
from pathlib import Path
from typing import Union

from ocrd_browser.model import Document
from ocrd_browser.util.config import _Tool
from ocrd_browser.util.launcher import Launcher
from tests import TestCase, TEST_BASE_PATH, data_provider

BASE_PATH = TEST_BASE_PATH / 'example/workspaces/heavy quoting/'


def quoted(file: Union[str, Path]) -> str:
    return shlex.quote(str(file))


def template_testcases():
    return [
        ('', ''),
        ('{ escaped brace }', '{{ escaped brace }}'),
        (quoted(BASE_PATH), '{workspace.directory}'),
        (quoted(BASE_PATH), '{workspace.baseurl}'),
        (quoted(BASE_PATH / 'mets.xml'), '{workspace.mets_target}'),
        ('--workspace=' + quoted(BASE_PATH), '--workspace={workspace.baseurl}'),
        ('application/vnd.prima.page+xml', '{file.mimetype}'),
        ('PHYS_0017', '{file.pageId}'),
        ('OCR-D-GT-PAGE', '{file.fileGrp}'),
        ('.xml', '{file.extension}'),
        ("'PAGE_0017 PÄGE.xml'", '{file.basename}'),
        ("'PAGE_0017 PÄGE'", '{file.basename_without_extension}'),
        ("'OCR-D-GT PÆGE/PAGE_0017 PÄGE.xml'", '{file.path.relative}'),
        (quoted(BASE_PATH / 'OCR-D-GT PÆGE/PAGE_0017 PÄGE.xml'), '{file.path.absolute}'),
        (quoted(BASE_PATH / 'OCR-D-GT PÆGE/PAGE_0017 PÄGE.xml'), '{file.path}')

    ]


class LauncherTestCase(TestCase):

    def setUp(self) -> None:
        self.launcher = Launcher()
        self.doc = Document.load(BASE_PATH / 'mets.xml')
        self.file = self.doc.files_for_page_id('PHYS_0017', 'OCR-D-GT-PAGE')[0]

    def test_launch_missing(self):
        setOverrideLogLevel('ERROR', True)
        with self.assertLogs('ocrd_browser.util.launcher.Launcher.launch', level='ERROR') as log_watch:
            process = self.launcher.launch('missingtool', self.doc, self.file)
        self.assertIsNone(process)
        self.assertEqual(3, len(log_watch.records))
        self.assertEqual(
            'Tool "missingtool" not found in your config, to fix place the following section in your ocrd-browser.conf',
            log_watch.records[0].getMessage()
        )
        setOverrideLogLevel('OFF', True)

    def test_launch_tool(self):
        tool = _Tool('Echotest', 'echo -n {file.path.relative}')
        with self.launcher.launch_tool(tool, self.doc, self.file, stdout=subprocess.PIPE) as process:
            process.wait()
            result = process.stdout.read().decode('utf8')
        self.assertEqual('OCR-D-GT PÆGE/PAGE_0017 PÄGE.xml', result)

    @data_provider(template_testcases)
    def test_template(self, expected: Union[Path, str], template: str):
        actual = self.launcher._template(template, self.doc, self.file)
        self.assertEqual(str(expected), actual)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import List, Optional, Any, Tuple, Dict, Union, cast, Set, TYPE_CHECKING

from PIL.Image import Image
from lxml.etree import ElementBase as Element
from inspect import signature
from deprecated import deprecated

from ocrd import Workspace, OcrdFile, OcrdExif
from ocrd_utils import MIMETYPE_PAGE, getLogger, pushd_popd, VERSION as OCRD_VERSION
from ocrd_models.ocrd_page import PcGtsType, PageType, MetadataType
from ocrd_models.constants import NAMESPACES

if TYPE_CHECKING:
    from ocrd_browser.model import Document

IMAGE_FROM_PAGE_FILENAME_SUPPORT = 'filename' in signature(Workspace.image_from_page).parameters


class Page:
    def __init__(self, document: Document, id_: str, file_group: str):
        self.document = document
        self._id = id_
        self.file_group = file_group
        self._pc_gts: Optional[PcGtsType] = None
        self._images: Optional[List[Image]] = None
        self._image_files: Optional[List[OcrdFile]] = None
        self._page_file: Optional[OcrdFile] = None

    @property
    def images(self) -> List[Image]:
        if self._images is None:
            self._images = [self.document.resolve_image(f) for f in self.image_files]
        return self._images

    @property
    def image_files(self) -> List[OcrdFile]:
        if self._image_files is None:
            self._image_files = self.get_files("//image/.*")
        return self._image_files

    def get_files(self, mimetype: str = MIMETYPE_PAGE) -> List[OcrdFile]:
        return self.document.files_for_page_id(self.id, self.file_group, mimetype=mimetype)

    @property
    def page_file(self) -> OcrdFile:
        if self._page_file is None:
            page_files = self.get_files(MIMETYPE_PAGE)
            if page_files:
                self._page_file = page_files[0]
        return self._page_file

    @property  # type: ignore[misc]
    @deprecated(reason="Makes no sense anymore, what is **the** file of a page? Use get_files() or page_file / image_files instead")
    def file(self) -> Optional[OcrdFile]:
        """
        TODO: Makes no sense anymore, what is **the** file of a page?

        The whole Page class needs to be split in "PageProxy" and PageXmlPage maybe
        @return:  Optional[OcrdFile]
        """
        if self.page_file:
            return self.page_file
        elif self.image_files:
            return next(iter(self.image_files))
        else:
            # noinspection PyTypeChecker
            any_files = self.get_files(mimetype=None)
            if any_files:
                return next(iter(any_files))

        return None

    @property
    def pc_gts(self) -> PcGtsType:
        if self._pc_gts is None:
            if self.page_file:
                self._pc_gts = self.document.page_for_file(self.page_file)
            else:
                image_files = self.image_files
                if len(image_files) > 0:
                    self._pc_gts = self.document.page_for_file(image_files[0])
        return self._pc_gts

    def get_image(self, feature_selector: Union[str, Set[str]] = '', feature_filter: Union[str, Set[str]] = '', filename: str = '') -> Tuple[Image, Dict[str, Any], OcrdExif]:
        log = getLogger('ocrd_browser.model.page.Page.get_image')

        ws: Workspace = self.document.workspace
        kwargs = {
            'transparency': True,
            'feature_selector': feature_selector if isinstance(feature_selector, str) else ','.join(sorted(feature_selector)),
            'feature_filter': feature_filter if isinstance(feature_filter, str) else ','.join(sorted(feature_filter))
        }
        if filename:
            if IMAGE_FROM_PAGE_FILENAME_SUPPORT:
                kwargs['filename'] = filename
            else:
                raise RuntimeError('Parameter filename not supported in ocrd version {}, at least 2.33.0 needed'.format(OCRD_VERSION))

        try:
            with pushd_popd(ws.directory):
                page_image, page_coords, page_image_info = ws.image_from_page(self.page, self.id, **kwargs)
        except Exception as e:
            log.exception(e)
            page_image, page_coords, page_image_info = None, None, None

        return page_image, page_coords, page_image_info

    @property
    def id(self) -> str:
        return self._id

    @property
    def page(self) -> PageType:
        return self.pc_gts.get_Page()

    @property
    def meta(self) -> MetadataType:
        return self.pc_gts.get_Metadata()

    def xpath(self, xpath: str) -> List[Element]:
        page_namespace = {'page': ns for ns in self.xml_root.nsmap.values() if ns.startswith('http://schema.primaresearch.org/PAGE/gts/pagecontent/')}
        return cast(List[Element], self.xml_root.xpath(xpath, namespaces=dict(NAMESPACES, **page_namespace)))

    @property
    def xml_root(self) -> Element:
        if self.pc_gts.gds_elementtree_node_ is None:
            from ocrd_models.ocrd_page_generateds import parsexmlstring_
            from io import StringIO
            sio = StringIO()

            self.pc_gts.export(
                outfile=sio,
                level=0,
                name_='PcGts',
                namespaceprefix_='pc:',
                namespacedef_='xmlns:pc="%s" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="%s %s/pagecontent.xsd"' % (
                    NAMESPACES['page'],
                    NAMESPACES['page'],
                    NAMESPACES['page']
                ))
            self.pc_gts.gds_elementtree_node_ = parsexmlstring_(sio.getvalue())  # pylint: disable=undefined-variable

        return self.pc_gts.gds_elementtree_node_



import re
from collections import Counter
from typing import List, Tuple, Optional, Counter as CounterType


def weight_match(s: str, preferreds: Optional[List[str]] = None) -> float:
    """
    Weights how good a string matches a list of regular expressions
    """
    weight = 0.0
    if preferreds:
        ln = float(len(preferreds))
        for i, preferred in enumerate(preferreds):
            if re.fullmatch(preferred, s):
                # prefer matches earlier in the list
                weight += (ln - i) / ln
                # break or no break???
    return weight


def best_file_group(file_groups_and_mimetypes: List[Tuple[str, str]], preferred_groups: Optional[List[str]] = None, preferred_mimetypes: Optional[List[str]] = None, cutoff: float = -9999) -> Optional[Tuple[str, str]]:
    file_groups: CounterType[Tuple[str, str]] = Counter()
    for file_group_and_mimetype in file_groups_and_mimetypes:
        file_group, mimetype = file_group_and_mimetype
        file_groups[file_group_and_mimetype] += weight_match(file_group, preferred_groups)  # type: ignore[assignment]
        file_groups[file_group_and_mimetype] += weight_match(mimetype, preferred_mimetypes)  # type: ignore[assignment]
        file_groups[file_group_and_mimetype] -= len(file_group) * 0.0001  # type: ignore[assignment]
    if file_groups:
        [(file_group_and_mimetype, score)] = file_groups.most_common(1)
        if score > cutoff:
            return file_group_and_mimetype
        else:
            return None
    else:
        return None



"""
Page-XML rendering object

This is heavily based on ocrd_segment.extract_pages (https://github.com/OCR-D/ocrd_segment/blob/master/ocrd_segment/extract_pages.py)
"""
from __future__ import annotations
from typing import Optional, Dict, Any, Union, List, Iterator, Tuple, Type, cast
import PIL.ImageFont
import numpy as np
from math import sin, cos, radians, inf
from enum import IntFlag
from collections import defaultdict
from logging import Logger

from functools import lru_cache as memoized

from PIL import ImageDraw, Image, ImageFont

from ocrd_models.ocrd_page import PcGtsType, PageType, BorderType, PrintSpaceType, RegionType, TextRegionType, TextLineType, WordType, GlyphType, GraphemeType, ChartRegionType, GraphicRegionType, SeparatorRegionType
from ocrd_utils import coordinates_of_segment, getLogger, polygon_from_points, transform_coordinates

from shapely.geometry import Polygon, Point, LineString
from shapely.validation import explain_validity
from shapely import prepared

RegionWithCoords = Union[RegionType, TextLineType, WordType, GlyphType, GraphemeType, PrintSpaceType, BorderType]
__all__ = ['PageXmlRenderer', 'RegionMap', 'Feature', 'Region']

CLASSES = {
    '': 'FFFFFF00',
    'Glyph': '2E8B08FF',
    'Word': 'B22222FF',
    'TextLine': '32CD32FF',
    'Baseline': '22DD227F',
    'Border': 'FFFFFFFF',
    'PrintSpace': 'CCCCCCFF',
    'TableRegion': '8B4513FF',
    'AdvertRegion': '4682B4FF',
    'ChemRegion': 'FF8C00FF',
    'MusicRegion': '9400D3FF',
    'MapRegion': '9ACDD2FF',
    'TextRegion': '0000FFFF',
    'TextRegion:paragraph': '0000FFFA',
    'TextRegion:heading': '0000FFF5',
    'TextRegion:caption': '0000FFF0',
    'TextRegion:header': '0000FFEB',
    'TextRegion:footer': '0000FFE6',
    'TextRegion:page-number': '0000FFE1',
    'TextRegion:drop-capital': '0000FFDC',
    'TextRegion:credit': '0000FFD7',
    'TextRegion:floating': '0000FFD2',
    'TextRegion:signature-mark': '0000FFCD',
    'TextRegion:catch-word': '0000FFC8',
    'TextRegion:marginalia': '0000FFC3',
    'TextRegion:footnote': '0000FFBE',
    'TextRegion:footnote-continued': '0000FFB9',
    'TextRegion:endnote': '0000FFB4',
    'TextRegion:TOC-entry': '0000FFAF',
    'TextRegion:list-label': '0000FFA5',
    'TextRegion:other': '0000FFA0',
    'ChartRegion': '800080FF',
    'ChartRegion:bar': '800080FA',
    'ChartRegion:line': '800080F5',
    'ChartRegion:pie': '800080F0',
    'ChartRegion:scatter': '800080EB',
    'ChartRegion:surface': '800080E6',
    'ChartRegion:other': '800080E1',
    'GraphicRegion': '008000FF',
    'GraphicRegion:logo': '008000FA',
    'GraphicRegion:letterhead': '008000F0',
    'GraphicRegion:decoration': '008000EB',
    'GraphicRegion:frame': '008000E6',
    'GraphicRegion:handwritten-annotation': '008000E1',
    'GraphicRegion:stamp': '008000DC',
    'GraphicRegion:signature': '008000D7',
    'GraphicRegion:barcode': '008000D2',
    'GraphicRegion:paper-grow': '008000CD',
    'GraphicRegion:punch-hole': '008000C8',
    'GraphicRegion:other': '008000C3',
    'ImageRegion': '00CED1FF',
    'LineDrawingRegion': 'B8860BFF',
    'MathsRegion': '00BFFFFF',
    'NoiseRegion': 'FF0000FF',
    'SeparatorRegion': 'FF00FFFF',
    'UnknownRegion': '646464FF',
    'CustomRegion': '637C81FF'
}


class Feature(IntFlag):
    IMAGE = 1
    BORDER = 2
    PRINT_SPACE = 4
    ORDER = 8
    REGIONS = 16
    LINES = 32
    WORDS = 64
    GLYPHS = 128
    GRAPHEMES = 256
    WARNINGS = 512
    BASELINES = 1024

    DEFAULT = 1 | 2 | 16 | 32 | 1024

    def should_render(self, region_ds: RegionWithCoords) -> bool:
        lookup = {
            self.BORDER: BorderType,
            self.PRINT_SPACE: PrintSpaceType,
            self.REGIONS: RegionType,
            self.LINES: TextLineType,
            self.WORDS: WordType,
            self.GLYPHS: GlyphType,
            # self.GRAPHEMES: GraphemeType
        }
        for feature, region_type in lookup.items():
            if feature & self and isinstance(region_ds, region_type):
                return True
        return False


class Region:
    """
    A wrapper around all types of Page-XML regions and their polygon with convenience methods
    """
    def __init__(self, region: RegionWithCoords) -> None:
        self.region = region
        self._poly: Optional[Polygon] = None
        self._prep_poly: Optional[prepared.PreparedGeometry] = None
        self.warnings: List[str] = []

    @property
    def coords_conf(self) -> Optional[float]:
        return cast(float, self.region.Coords.conf) if hasattr(self.region, 'Coords') else None

    @property
    def text_conf(self) -> Optional[float]:
        if isinstance(self.region, (TextRegionType, TextLineType, WordType, GlyphType)):
            if self.region.get_TextEquiv() and self.region.get_TextEquiv()[0].conf:
                return cast(float, self.region.get_TextEquiv()[0].conf)
        return None

    @property
    def poly(self) -> Polygon:
        return self._poly

    @poly.setter
    def poly(self, poly: Polygon) -> None:
        self._poly = poly
        self._prep_poly = prepared.prep(self._poly)

    def contains(self, p: Point) -> bool:
        return self._prep_poly and cast(bool, self._prep_poly.contains(p))

    @property
    def id(self) -> str:
        if hasattr(self.region, 'id'):
            return str(self.region.id)
        elif hasattr(self.region, 'pcGtsId'):
            return str(self.region.pcGtsId)
        elif hasattr(self.region, 'imageFilename'):
            return str(self.region.imageFilename)
        else:
            return ''

    @property
    def region_type(self) -> str:
        return self.base_type + (':' + self.region_subtype if self.region_subtype else '')

    @property
    def base_type(self) -> str:
        return cast(str, self.region.__class__.__name__[:-4])

    @property
    def region_subtype(self) -> str:
        return cast(str, self.region.get_type()) if isinstance(self.region, (TextRegionType, ChartRegionType, GraphicRegionType)) else ''

    @property
    def text(self) -> str:
        if isinstance(self.region, (TextRegionType, TextLineType, WordType, GlyphType)):
            if self.region.get_TextEquiv() and self.region.get_TextEquiv()[0].Unicode:
                return cast(str, self.region.get_TextEquiv()[0].Unicode)
        return ''

    @property
    def parent(self) -> Optional['Region']:
        return Region(self.region.parent_object_) if hasattr(self.region, 'parent_object_') and self.region.parent_object_ else None

    @memoized(maxsize=1)
    def depth(self) -> int:
        return len(self.breadcrumbs())

    @memoized(maxsize=1)
    def breadcrumbs(self) -> List['Region']:
        """
        Traverses region up to the root (PcGts) element
        """
        breadcrumbs: List[Region] = []
        r = self
        while r:
            breadcrumbs.append(r)
            r = r.parent
        return list(breadcrumbs)

    def __str__(self) -> str:
        return '{:s}{:s}'.format(self.region_type, '#' + self.id)

    def __hash__(self) -> int:
        return id(self.region)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.region is other.region


class RegionBase:
    def __init__(self) -> None:
        self.children: List[RegionNode] = []

    def append(self, node: RegionNode) -> None:
        self.children.append(node)

    def find_region(self, x: float, y: float, ignore_regions: Optional[List[Type[RegionWithCoords]]] = None) -> Optional[Region]:
        """
        Finds deepest region at x,y
        """
        ignore_regions = ignore_regions if ignore_regions is not None else [BorderType, PrintSpaceType]

        def filter_regions(n: Region) -> bool:
            return type(n.region) not in ignore_regions

        p = Point(x, y)

        regions = list(filter(filter_regions, self.find_regions(p)))
        return regions[-1] if regions else None

    def find_regions(self, p: Point) -> List[Region]:
        regions: List[Region] = []
        for node in self.children:
            if node.region.contains(p):
                regions.append(node.region)
                regions.extend(node.find_regions(p))
        return regions


class RegionNode(RegionBase):
    def __init__(self, region: Region):
        super().__init__()
        self.region = region

    def __str__(self) -> str:
        return '{}'.format(self.region)


class RegionMap(RegionBase):
    """
    Builds a tree of regions based on their Page-XML nesting.
    """
    def __init__(self) -> None:
        super().__init__()
        self.nodes_by_region: Dict[Region, RegionNode] = {}
        self.region_by_id: Dict[str, Region] = {}

    def refetch(self, r: Region) -> Optional[Region]:
        return self.get(r.id) if r else None

    def get(self, id_: str) -> Optional[Region]:
        return self.region_by_id.get(id_, None)

    def append(self, node: RegionNode) -> None:
        """
        appends a RegionNode to the tree
        IMPORTANT: this will only work if the parent node already exists, so the calls have to be in breadth-first or depth-first pre-order order.
        The Operations class will iterate in breadth-first order
        """
        self.nodes_by_region[node.region] = node
        if node.region.id:
            self.region_by_id[node.region.id] = node.region

        if node.region.parent in self.nodes_by_region:
            parent_region = self.nodes_by_region[node.region.parent]
            parent_region.append(node)
        else:
            self.children.append(node)


class Operation:
    """
    Base class for a rendering Operation in a certain depth (or layer)
    An Operation will render itself and/or add itself to the RegionMap in paint()
    """
    def __init__(self, color: str, depth: int):
        self.color = color
        self.depth = depth

    def paint(self, draw: ImageDraw.Draw, regions: RegionMap) -> None:
        pass


class PolygonOperation(Operation):

    def __init__(self, region: Region, fill: str, outline: str):
        super().__init__(outline, region.depth() * 10)
        self.region = region
        self.fill = fill

    def paint(self, draw: ImageDraw.Draw, regions: RegionMap) -> None:
        xy = list(map(tuple, self.region.poly.exterior.coords[:-1]))
        draw.polygon(xy, self.fill, self.color)
        regions.append(RegionNode(self.region))


class LineStringOperation(Operation):
    def __init__(self, linestring: LineString, color: str, width: int = 4):
        super().__init__(color, 45)  # Depth 45 is between 40(TextLine) and 50(Word)
        self.linestring = linestring
        self.width = width

    def paint(self, draw: ImageDraw.Draw, regions: RegionMap) -> None:
        xy = list(map(tuple, self.linestring.coords))
        draw.line(xy, self.color, self.width)


class ArrowOperation(Operation):
    def __init__(self, p0: Point, p1: Point, size: float = 30.0, width: int = 3, color: str = '#FF0000FF'):
        super().__init__(color, 200)  # Depth 200 is on top of everything
        self.p0 = p0.coords[0]
        self.p1 = p1.coords[0]
        self.size = size
        self.width = width

    def paint(self, draw: ImageDraw.Draw, regions: RegionMap) -> None:
        angle = radians(180.0 - 30)  # 30 degrees
        c, s = cos(angle), sin(angle)
        d = self.p1[0] - self.p0[0], self.p1[1] - self.p0[1]
        left = d[0] * c - d[1] * s, d[0] * s + d[1] * c
        right = d[0] * c + d[1] * s, -d[0] * s + d[1] * c
        lf = self.size / (d[0] ** 2 + d[1] ** 2) ** 0.5

        # Draw arrow shaft
        draw.line([self.p0, self.p1], fill=self.color, width=self.width)
        # Draw dot
        draw.ellipse((self.p1[0] - 5, self.p1[1] - 5, self.p1[0] + 5, self.p1[1] + 5), fill=self.color)
        # Draw left arrow wing
        draw.line([(self.p1[0] + lf * left[0], self.p1[1] + lf * left[1]), self.p1], fill=self.color, width=self.width)
        # Draw right arrow wing
        draw.line([(self.p1[0] + lf * right[0], self.p1[1] + lf * right[1]), self.p1], fill=self.color, width=self.width)


class TextOperation(Operation):

    font_cache: Dict[Tuple[str, int], ImageFont.FreeTypeFont] = {}

    @classmethod
    def font(cls, file: str, size: int) -> ImageFont.FreeTypeFont:
        if (file, size) not in cls.font_cache:
            cls.font_cache[(file, size)] = ImageFont.truetype(file, size=size, layout_engine=PIL.ImageFont.LAYOUT_RAQM)
        return cls.font_cache[(file, size)]

    def __init__(self, region: Region, color: str):
        super().__init__(color, 55)  # Depth 45 is between 40(TextLine) and 50(Word)
        self.region = region

    def paint(self, draw: ImageDraw.Draw, regions: RegionMap) -> None:
        sx, sy, ex, ey = self.region.poly.bounds
        font_size = 30
        factor = 0
        tries = 0

        while abs(factor - 1) > 0.05 and tries < 5:
            font = self.font('/usr/share/fonts/truetype/ubuntu/Ubuntu-R.ttf', font_size)
            w, h = draw.textsize(self.region.text, font=font)
            # TODO super random compromise between fit to width and fit to height
            factor = 0.7 * (ex - sx) / w + 0.3 * (ey - sy) / h
            font_size = int(font_size * factor)
            tries += 1

        draw.text((sx, sy), self.region.text, fill=self.color, font=font, anchor="lt")


class Operations:
    """
    Operations is a depth-sorted List of Operation objects

    Each depth can be plotted on its own image-layer and will be blended with Image.alpha_composite, so
    Image.alpha_composite will only be called once per layer instead of once per operation
    """

    def __init__(self) -> None:
        self.operations: Dict[int, List[Operation]] = defaultdict(list)

    def append(self, op: Operation) -> None:
        self.operations[op.depth].append(op)

    def layers(self) -> Iterator[Tuple[int, List[Operation]]]:
        for layer in sorted(self.operations, reverse=False):
            yield layer, self.operations[layer]

    def paint(self, canvas: Image.Image) -> Tuple[Image.Image, RegionMap]:
        """
        Paints the operations on canvas and fills the RegionMap accordingly
        """
        regions = RegionMap()
        for depth, operations in self.layers():
            layer = Image.new(mode='RGBA', size=canvas.size, color='#FFFFFF00')
            draw = ImageDraw.Draw(layer)
            for operation in operations:
                operation.paint(draw, regions)
            canvas.alpha_composite(layer)
        self.operations.clear()
        return canvas, regions


class RegionFactory:
    def __init__(self, coords: Dict[str, Any], page_id: str = '<unknown>', logger: Logger = None):
        self.coords = coords
        self.page_id = page_id
        self.logger = logger or getLogger(self.__class__.__module__ + '.' + self.__class__.__name__)

    def create(self, region_ds: RegionWithCoords) -> Optional[Region]:
        if not region_ds:
            return None

        region = Region(region_ds)
        coords = coordinates_of_segment(region_ds, None, self.coords)

        warnings = []

        try:
            poly = Polygon(coords)
        except ValueError as err:
            self.logger.error('Page "%s" @ %s %s', self.page_id, str(region), str(err))
            return None

        if not poly.is_valid:
            warning = explain_validity(poly)
            poly, error = self.make_valid(poly)
            if not poly.is_valid:
                self.logger.error('Page "%s" @ %s %s', self.page_id, str(region), str(warning))
                return None
            else:
                warnings.append('{} fixed with an error of {:.3%}'.format(warning, error))

        if poly.length < 4:
            warnings.append(str('has too few points'))

        if poly.is_empty or not poly.area > 0:
            self.logger.error('Page "%s" @ %s %s', self.page_id, str(region), 'is empty')
            return None

        if poly.bounds[0] < 0 or poly.bounds[1] < 0:
            warnings.append('is negative')

        if warnings:
            self.logger.warning('Page "%s" @ %s %s', self.page_id, str(region), ' | '.join(warnings))

        region.poly = poly
        region.warnings = warnings

        return region

    def create_baseline(self, text_line: TextLineType) -> Optional[LineString]:
        if text_line.get_Baseline() is None or text_line.get_Baseline().points is None:
            return None

        points = np.array(polygon_from_points(text_line.get_Baseline().points))
        points = transform_coordinates(points, self.coords['transform'])
        try:
            line = LineString(np.round(points).astype(np.int32))
        except ValueError as err:
            self.logger.error('Page "%s" @ %s/Baseline %s', self.page_id, str(text_line.id), str(err))
            return None

        if not line.is_valid:
            self.logger.error('Page "%s" @ %s/Baseline %s', self.page_id, str(text_line.id), str(explain_validity(line)))
            return None

        return line

    @staticmethod
    def make_valid(polygon: Polygon) -> Tuple[Polygon, float]:
        """Ensures shapely.geometry.Polygon object is valid by repeated simplification"""
        tolerance = 1
        for split in range(1, len(polygon.exterior.coords) - 1):
            if polygon.is_valid or polygon.simplify(polygon.area).is_valid:
                break
            # simplification may not be possible (at all) due to ordering
            # in that case, try another starting point
            polygon = Polygon(polygon.exterior.coords[-split:] + polygon.exterior.coords[:-split])
        for tolerance in range(1, int(polygon.area)):
            if polygon.is_valid:
                break
            # simplification may require a larger tolerance
            polygon = polygon.simplify(tolerance)

        a = polygon.area
        return polygon, tolerance / a if a > 0 else inf


class PageXmlRenderer:

    def __init__(self, canvas: Image.Image, coords: Dict[str, Any], page_id: str = '<unknown>',
                 features: Optional[Feature] = None, colors: Optional[Dict[str, str]] = None, logger: Logger = None):
        self.features = features or Feature.DEFAULT

        if self.features & Feature.IMAGE:
            self.canvas = canvas.convert('RGBA')
        else:
            self.canvas = Image.new(mode='RGBA', size=canvas.size, color='#FFFFFFFF')

        self.region_factory = RegionFactory(coords, page_id, logger)

        self.colors: Dict[str, str] = defaultdict(lambda: 'FF0000FF')
        self.colors.update(colors or CLASSES)

        self.operations = Operations()

    def render_all(self, pc_gts: PcGtsType) -> None:
        page: PageType = pc_gts.get_Page()
        self.render_type(page.get_PrintSpace())
        self.render_type(page.get_Border())

        def region_priority(region: RegionType) -> int:
            # often, regions overlap; since we don't alpha-composite,
            # we should provide a useful default priority of what is
            # rendered last (and thus unoccluded); document order is
            # not helpful, because it depends on the workflow or geometry
            # instead of visibility concerns; thus, ensure that:
            # - SeparatorRegion is top-most
            # - TextRegion is next
            # - all others are next
            if isinstance(region, SeparatorRegionType):
                return 0
            if isinstance(region, TextRegionType):
                return -1
            return -2
        for region_ds in sorted(page.get_AllRegions(), key=region_priority):
            self.render_type(region_ds)

        if self.features & Feature.ORDER:
            last_point: Optional[Point] = None
            for region_ds in page.get_AllRegions(order='reading-order-only'):
                region = self.region_factory.create(region_ds)
                new_point = region.poly.representative_point()
                if last_point:
                    self.operations.append(ArrowOperation(last_point, new_point, color='#FF0000CF'))
                last_point = new_point

    def get_result(self) -> Tuple[Image.Image, RegionMap]:
        canvas, regions = self.operations.paint(self.canvas.copy())
        return canvas, regions

    def render_type(self, region_ds: RegionWithCoords) -> None:
        if self.features & Feature.BASELINES and isinstance(region_ds, TextLineType):
            linestring = self.region_factory.create_baseline(region_ds)
            if linestring:
                self.operations.append(LineStringOperation(linestring, '#' + self.colors['Baseline']))

        if self.features.should_render(region_ds):
            region = self.region_factory.create(region_ds)
            if region:
                # if isinstance(region_ds, WordType):
                #    self.operations.append(TextOperation(region, '#000000FF'))

                if self.features & Feature.WARNINGS and region.warnings:
                    op = PolygonOperation(region, '#FF00003E', '#FF000076')
                else:
                    color = self.colors[region.region_type]
                    op = PolygonOperation(region, '#' + color[:6] + '1E', '#' + color[:6] + '96')
                self.operations.append(op)

        if isinstance(region_ds, TextRegionType):
            self.render_text_region(region_ds)

    def render_text_region(self, text_region: TextRegionType) -> None:
        line: TextLineType
        word: WordType
        glyph: GlyphType
        for line in text_region.get_TextLine():
            self.render_type(line)
            for word in line.get_Word():
                self.render_type(word)
                for glyph in word.get_Glyph():
                    self.render_type(glyph)



from tempfile import mkdtemp
from os import makedirs, environ
from os.path import isdir, exists, abspath, dirname, realpath
from shutil import copytree, rmtree
from contextlib import contextmanager

@contextmanager
def copy_of_directory(src, dst=None):
    if not dst:
        dst = mkdtemp(prefix='ocrd-test-tempdir-')
    if isdir(dst):
        rmtree(dst)
    if not isdir(dirname(dst)):
        makedirs(dirname(dst))
    copytree(src, dst, symlinks=True)
    yield dst
    rmtree(dst)


class Assets():
    """
    Access test assets in OCR-D/assets
    """

    def __init__(self, baseurl=None):
        self.baseurl = baseurl

    def url_of(self, path, baseurl=None):
        """
        Get the URL of an asset
        """
        if baseurl is None:
            baseurl = self.baseurl
        return self.baseurl + path

    def path_to(self, path, baseurl=None):
        return self.url_of(path, baseurl)

if 'OCRD_BASEURL' in environ:
    assets = Assets(environ['OCRD_BASEURL'])
else:
    assets = Assets(dirname(realpath(__file__)) + '/')



import unittest
from PIL import Image, ImageDraw
from gi.repository import GdkPixbuf

from tests import TestCase, ASSETS_PATH, data_provider
from ocrd_browser.util.image import pil_to_pixbuf, _pil_to_pixbuf_via_cv, _pil_to_pixbuf_via_pixbuf_loader
from timeit import timeit


def _image_modes():
    return (
        ('RGB',
         (0, 0, 0), (127, 127, 127), (255, 255, 255),
         (0, 0, 0), (127, 127, 127), (255, 255, 255)
         ),
        ('RGBA',
         (0, 0, 0, 128), (255, 0, 255, 255), (255, 255, 255, 255),
         (0, 0, 0, 128), (255, 0, 255, 255), (255, 255, 255, 255)
         ),
        ('1',
         (0,), (1,), (1,),
         (0, 0, 0), (255, 255, 255), (255, 255, 255)
         ),
        ('LA',
         (0, 127), (255, 255), (0, 127),
         (0, 0, 0, 127), (255, 255, 255, 255), (0, 0, 0, 127)
         ),
    )


class ImageUtilTestCase(TestCase):

    def test_pil_to_pixbuf_is_faster_via_opencv(self):
        self.skipTest('Slow test')
        files = [
            ASSETS_PATH / 'kant_aufklaerung_1784-binarized/data/OCR-D-IMG/OCR-D-IMG_0017.tif',
            ASSETS_PATH / 'kant_aufklaerung_1784-binarized/data/OCR-D-IMG-1BIT/OCR-D-IMG-1BIT_0017.png',
            ASSETS_PATH / 'kant_aufklaerung_1784-binarized/data/OCR-D-IMG-NRM/OCR-D-IMG-NRM_0017.png',
        ]
        for file in files:
            via_cv_time, via_pil_time = self._test_pil_to_pixbuf_performance_on_file(file)
            print('OpenCV({}s) vs. PIL({}s): {} for {}'.format(via_cv_time, via_pil_time, via_cv_time / via_pil_time,
                                                               file))
            self.assertLess(via_cv_time, via_pil_time, 'via_cv took longer for {}'.format(file))

    @data_provider(_image_modes)
    def test_image_modes(self, mode, bg, fg1, fg2, bg_test, fg1_test, fg2_test):
        pil = self._generate_test_image(mode, bg, fg1, fg2)
        pb = pil_to_pixbuf(pil)
        self.assertSequenceEqual(bg_test, self._get_pixbuf_pixel(pb, 0, 10))
        self.assertSequenceEqual(fg1_test, self._get_pixbuf_pixel(pb, 0, 0))
        self.assertSequenceEqual(fg2_test, self._get_pixbuf_pixel(pb, pil.size[0] - 1, 1))

    @staticmethod
    def _get_pixbuf_pixel(pb: GdkPixbuf.Pixbuf, x, y):
        bytes = pb.get_pixels()
        n_channels = pb.get_n_channels()
        rowstride = pb.get_rowstride()
        bps = pb.get_bits_per_sample()

        bytes_per_pixel = int(n_channels * bps / 8)
        offset = y * rowstride + x * bytes_per_pixel

        return list(bytes[offset:offset + bytes_per_pixel])

    @staticmethod
    def _generate_test_image(mode, bg, fg1, fg2):
        im = Image.new(mode, (50, 30), bg)
        draw = ImageDraw.Draw(im)
        w, h = im.size
        draw.line((0, 0, w, h), fill=fg1)
        draw.line((0, h, w, 0), fill=fg2)
        return im

    @staticmethod
    def _test_pil_to_pixbuf_performance_on_file(file, number: int = 10):
        image: Image.Image = Image.open(file)
        image.load()

        via_cv_time = timeit(lambda: _pil_to_pixbuf_via_cv(image), number=number)
        via_pil_time = timeit(lambda: _pil_to_pixbuf_via_pixbuf_loader(image), number=number)

        return via_cv_time, via_pil_time


if __name__ == '__main__':
    unittest.main()



import unittest

from tests import TestCase
from ocrd_browser.util.file_groups import best_file_group, weight_match


class FileGroupsTestCase(TestCase):

    def setUp(self) -> None:
        self.test_groups = [
            ('OCR-D-SEG-BLOCK-tesseract', 'application/vnd.prima.page+xml'),
            ('OCR-D-GT-PAGE', 'application/vnd.prima.page+xml'),
            ('OCR-D-IMG', 'image/tiff'),
            ('OCR-D-IMG-BIN', 'image/tiff'),
        ]

    def test_weight_match(self):
        self.assertEqual(0, weight_match('ABC', []))
        self.assertEqual(1, weight_match('ABC', ['ABC']))
        self.assertEqual(1, weight_match('ABC', ['ABC', 'DEF']))
        self.assertEqual(0.5, weight_match('DEF', ['ABC', 'DEF']))
        self.assertEqual(0, weight_match('DEF', ['ABC']))

    def test_weight_match_re(self):
        self.assertEqual(1, weight_match('ABC', [r'AB[CD]']))
        # 1.0 for match ABC + 0.5 match \w{3}
        self.assertEqual(1.5, weight_match('ABC', ['ABC', r'\w{3}']))

    def test_best_file_group_prefers_shortest_file_group_if_no_other_preferences(self):
        expected = ('OCR-D-IMG', 'image/tiff')
        actual = best_file_group(self.test_groups)
        self.assertEqual(expected, actual)

    def test_best_file_group_prefers_by_mimetype(self):
        expected = ('OCR-D-GT-PAGE', 'application/vnd.prima.page+xml')
        actual = best_file_group(self.test_groups, preferred_mimetypes=[r'application/.*'])
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import MagicMock
from ocrd_browser.view import ViewPage
from ocrd_browser.ui import MainWindow
from tests import TestCase


class ViewPageTestCase(TestCase):

    def setUp(self):
        self.vx = ViewPage('unique', MagicMock(spec=MainWindow))

    def test_can_construct(self):
        self.assertIsNotNone(self.vx)


if __name__ == '__main__':
    unittest.main()
